import unittest
from unittest.mock import MagicMock
from unittest.mock import Mock

import pytest
from test_providers import chestnut
from test_providers import geo
from test_providers import larch
from test_providers import species
from test_providers import trees

from skosprovider.registry import Registry
from skosprovider.registry import RegistryException


class RegistryTests(unittest.TestCase):
    def setUp(self):
        self.reg = Registry()
        self.prov = trees
        self.prov2 = geo

    def tearDown(self):
        self.reg = None
        self.prov = None
        self.prov2 = None

    def test_default_metadata_is_dict(self):
        self.assertIsInstance(self.reg.get_metadata(), dict)

    def test_passed_metadata_is_dict(self):
        self.reg = Registry(metadata={"catalog": {"uri": "http://my.data.org"}})
        assert "catalog" in self.reg.get_metadata()
        assert "uri" in self.reg.get_metadata().get("catalog")

    def test_set_instance_scope(self):
        self.reg = Registry(instance_scope="threaded_global")
        assert self.reg.instance_scope == "threaded_global"

    def test_set_invalid_instance_scope(self):
        with pytest.raises(ValueError):
            Registry(instance_scope="bad_scope")

    def test_empty_register_provider(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_provider("TREES"), self.prov)

    def test_empty_register_removeProvider(self):
        self.assertFalse(self.reg.remove_provider("TREES"))

    def test_empty_getProviders(self):
        self.assertEqual(self.reg.get_providers(), [])
        self.assertEqual(self.reg.get_providers(ids=[]), [])

    def test_empty_getProviderById(self):
        self.assertFalse(self.reg.get_provider("TREES"))
        self.assertFalse(self.reg.get_provider("http://id.trees.org"))

    def test_empty_findConcepts(self):
        self.assertEqual(self.reg.find({}), [])

    def test_empty_getAllConcepts(self):
        self.assertEqual(self.reg.get_all(), [])

    def test_one_provider_register_provider(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_provider("TREES"), self.prov)
        self.assertEqual(self.reg.get_provider("http://id.trees.org"), self.prov)
        self.reg.register_provider(self.prov2)
        self.assertEqual(self.reg.get_provider("GEOGRAPHY"), self.prov2)
        self.assertEqual(
            self.reg.get_provider("urn:x-skosprovider:geography"), self.prov2
        )

    def test_one_provider_register_double_provider(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_provider("TREES"), self.prov)
        self.assertRaises(RegistryException, self.reg.register_provider, self.prov)
        # Change the id, but keep identical URIs
        self.prov.metadata["id"] = "TREESTOO"
        self.assertRaises(RegistryException, self.reg.register_provider, self.prov)
        self.prov.metadata["id"] = "TREES"

    def test_register_provider_wrong_scope(self):
        from skosprovider.skos import ConceptScheme
        from skosprovider.providers import DictionaryProvider

        trees = DictionaryProvider(
            {"id": "TREES", "default_language": "nl"},
            [larch, chestnut, species],
            concept_scheme=ConceptScheme("urn:something"),
            allowed_instance_scopes=["threaded_thread"],
        )
        with pytest.raises(RegistryException):
            self.reg.register_provider(trees)

    def test_register_provider_no_uri(self):
        p = Mock()
        p.allowed_instance_scopes = ["single"]
        p.get_vocabulary_id = MagicMock(return_value="MYID")
        del p.get_vocabulary_uri
        p.concept_scheme = MagicMock()
        p.concept_scheme.uri = "http://my.id.org"
        self.reg.register_provider(p)
        assert p == self.reg.get_provider("MYID")
        assert p == self.reg.get_provider("http://my.id.org")

    def test_remove_provider_no_uri(self):
        provider = Mock()
        provider.allowed_instance_scopes = ["single"]
        provider.get_vocabulary_id = MagicMock(return_value="MYID")
        del provider.get_vocabulary_uri
        provider.concept_scheme = MagicMock()
        provider.concept_scheme.uri = "http://my.id.org"
        self.reg.register_provider(provider)
        assert provider == self.reg.get_provider("MYID")
        assert provider == self.reg.get_provider("http://my.id.org")
        self.reg.remove_provider("http://my.id.org")
        assert not self.reg.get_provider("MYID")

    def test_one_provider_removeProvider(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_provider("TREES"), self.prov)
        self.reg.remove_provider("TREES")
        self.assertFalse(self.reg.get_provider("TREES"))

    def test_one_provider_removeProviderWithUri(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_provider("TREES"), self.prov)
        self.reg.remove_provider("http://id.trees.org")
        self.assertFalse(self.reg.get_provider("TREES"))

    def test_one_provider_getProviders(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_providers(), [self.prov])
        self.assertEqual(self.reg.get_providers(ids=["TREES"]), [self.prov])

    def test_one_provider_getProvidersWithIds(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_providers(ids=["TREES"]), [self.prov])
        self.assertEqual(self.reg.get_providers(), [self.prov])
        self.assertEqual(self.reg.get_providers(ids=["GEOGRAPHY"]), [])

    def test_one_provider_getProvidersWithUris(self):
        self.reg.register_provider(self.prov)
        assert self.reg.get_providers(ids=["http://id.trees.org"]) == [self.prov]
        assert self.reg.get_providers() == [self.prov]
        assert self.reg.get_providers(ids=["urn:x-skosprovider:geography"]) == []

    def test_one_provider_getProvidersWithSubject(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_providers(subject="something"), [])
        self.assertEqual(self.reg.get_providers(subject="biology"), [self.prov])

    def test_one_provider_getPoviderWithId(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_provider("TREES"), self.prov)

    def test_one_provider_getPoviderWithUri(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(self.reg.get_provider("http://id.trees.org"), self.prov)

    def test_one_provider_findConcepts(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(
            self.reg.find({"label": "The Larch"}),
            [
                {
                    "id": "TREES",
                    "concepts": [
                        {
                            "id": "1",
                            "uri": "http://id.trees.org/1",
                            "type": "concept",
                            "label": "De Lariks",
                        }
                    ],
                }
            ],
        )

    def test_one_provider_getConceptByUri(self):
        self.reg.register_provider(self.prov)
        c = self.reg.get_by_uri("http://id.trees.org/1")
        self.assertEqual(c.id, "1")
        self.assertEqual(c.uri, "http://id.trees.org/1")

    def test_one_provider_getConceptByUriDifferentFromConceptScheme(self):
        from skosprovider.skos import ConceptScheme
        from skosprovider.providers import DictionaryProvider

        trees = DictionaryProvider(
            {"id": "TREES", "default_language": "nl"},
            [larch, chestnut, species],
            concept_scheme=ConceptScheme("urn:something"),
        )
        self.reg.register_provider(trees)
        c = self.reg.get_by_uri("http://id.trees.org/1")
        self.assertEqual(c.id, "1")
        self.assertEqual(c.uri, "http://id.trees.org/1")

    def test_one_provider_getConceptByUnexistingUri(self):
        self.reg.register_provider(self.prov)
        c = self.reg.get_by_uri("http://id.thingy.com/123456")
        self.assertFalse(c)

    def test_get_by_invalid_uri(self):
        self.assertRaises(ValueError, self.reg.get_by_uri, None)

    def test_one_provider_findConceptsWithProviderid(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(
            self.reg.find({"label": "The Larch"}, providers=["TREES"]),
            [
                {
                    "id": "TREES",
                    "concepts": [
                        {
                            "id": "1",
                            "uri": "http://id.trees.org/1",
                            "type": "concept",
                            "label": "De Lariks",
                        }
                    ],
                }
            ],
        )
        self.assertEqual(self.reg.find({"label": "The Larch"}, providers=[]), [])

    def test_one_provider_getAllConcepts(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(
            self.reg.get_all(),
            [
                {
                    "id": "TREES",
                    "concepts": [
                        {
                            "id": "1",
                            "uri": "http://id.trees.org/1",
                            "type": "concept",
                            "label": "De Lariks",
                        },
                        {
                            "id": "2",
                            "uri": "http://id.trees.org/2",
                            "type": "concept",
                            "label": "De Paardekastanje",
                        },
                        {
                            "id": 3,
                            "uri": "http://id.trees.org/3",
                            "type": "collection",
                            "label": "Bomen per soort",
                        },
                    ],
                }
            ],
        )

    def test_one_provider_getAllConceptsDifferentLanguage(self):
        self.reg.register_provider(self.prov)
        self.assertEqual(
            self.reg.get_all(language="en"),
            [
                {
                    "id": "TREES",
                    "concepts": [
                        {
                            "id": "1",
                            "uri": "http://id.trees.org/1",
                            "type": "concept",
                            "label": "The Larch",
                        },
                        {
                            "id": "2",
                            "uri": "http://id.trees.org/2",
                            "type": "concept",
                            "label": "The Chestnut",
                        },
                        {
                            "id": 3,
                            "uri": "http://id.trees.org/3",
                            "type": "collection",
                            "label": "Trees by species",
                        },
                    ],
                }
            ],
        )

    def test_two_providers_findConceptsWithProviderIdAndUri(self):
        self.reg.register_provider(self.prov2)
        self.reg.register_provider(self.prov)
        self.assertEqual(
            self.reg.find({"label": "The Larch"}, providers=["TREES"]),
            self.reg.find({"label": "The Larch"}, providers=["http://id.trees.org"]),
        )

    def test_two_providers_findConceptsProvidersDictionarySyntax(self):
        self.reg.register_provider(self.prov2)
        self.reg.register_provider(self.prov)
        self.assertEqual(
            self.reg.find({"label": "The Larch"}, providers=["TREES"]),
            self.reg.find(
                {"label": "The Larch"}, providers={"ids": ["http://id.trees.org"]}
            ),
        )

    def test_one_provider_findConceptsWithSubject(self):
        self.reg.register_provider(self.prov)
        provs = self.reg.get_providers(subject="biology")
        res = [{"id": p.get_vocabulary_id(), "concepts": p.find({})} for p in provs]
        self.assertEqual(res, self.reg.find({}, subject="biology"))

    def test_one_provider_findConceptsWithSubject_language_en(self):
        self.reg.register_provider(self.prov)
        provs = self.reg.get_providers(subject="biology")
        res = [
            {"id": p.get_vocabulary_id(), "concepts": p.find({}, language="en")}
            for p in provs
        ]
        self.assertEqual(res, self.reg.find({}, subject="biology", language="en"))

    def test_one_provider_findConceptsWithSubject_language_nl(self):
        self.reg.register_provider(self.prov)
        provs = self.reg.get_providers(subject="biology")
        res = [{"id": p.get_vocabulary_id(), "concepts": p.find({})} for p in provs]
        self.assertEqual(res, self.reg.find({}, subject="biology", language="nl"))



import pytest

from skosprovider.skos import Collection
from skosprovider.skos import Concept
from skosprovider.skos import ConceptScheme
from skosprovider.skos import Label
from skosprovider.skos import Note
from skosprovider.skos import Source
from skosprovider.skos import dict_to_label
from skosprovider.skos import dict_to_note
from skosprovider.skos import dict_to_source
from skosprovider.skos import filter_labels_by_language
from skosprovider.skos import find_best_label_for_type
from skosprovider.skos import label


class TestLabel:

    def testConstructor(self):
        label = Label("Knokke-Heist", type="prefLabel", language="nl-BE")
        assert label.label == "Knokke-Heist"
        assert label.type == "prefLabel"
        assert label.language == "nl-BE"
        assert label.uri is None

    def testConstructorInvalidLanguage(self):
        with pytest.raises(ValueError):
            Label("Knokke-Heist", type="prefLabel", language="nederlands")
        label = Label("Knokke-Heist", type="prefLabel", language=None)
        assert label.language == "und"

    def testConstructorOptionalFields(self):
        label = Label(
            "Knokke-Heist",
            type="prefLabel",
            language="nl-BE",
            uri="urn:x-skosprovider:gemeenten:knokke-heist:nl-BE",
        )
        assert label.uri == "urn:x-skosprovider:gemeenten:knokke-heist:nl-BE"
        label = Label(
            "Knokke-Heist", uri="urn:x-skosprovider:gemeenten:knokke-heist:nl-BE"
        )
        assert label.uri == "urn:x-skosprovider:gemeenten:knokke-heist:nl-BE"

    def testRepr(self):
        label = Label("Knokke-Heist", type="prefLabel", language="nl-BE")
        assert repr(label) == "Label('Knokke-Heist', 'prefLabel', 'nl-BE')"
        label = Label(
            "Knokke-Heist",
            type="prefLabel",
            language="nl-BE",
            uri="urn:x-skosp:gem:KH:nl-Be",
        )
        assert (
            repr(label)
            == "Label('Knokke-Heist', 'prefLabel', 'nl-BE', 'urn:x-skosp:gem:KH:nl-Be')"
        )

    def testIsValidType(self):
        assert Label.is_valid_type("prefLabel")
        assert not Label.is_valid_type("voorkeursLabel")
        label = Label("Knokke-Heist")
        assert label.is_valid_type("prefLabel")

    def testEquality(self):
        label1 = Label("Knokke-Heist")
        label2 = Label("Knokke-Heist", "prefLabel", "und")
        assert label1 == label2

    def testInequality(self):
        label1 = Label("Knokke-Heist")
        label2 = Label("Knokke", "altLabel")
        assert label1 != label2

    def testDictEquality(self):
        label1 = Label("Knokke-Heist")
        label2 = {"label": "Knokke-Heist", "type": "prefLabel", "language": "und"}
        assert label1 == label2

    def testDictInequality(self):
        label1 = Label("Knokke-Heist")
        label2 = {"label": "Knokke", "type": "altLabel", "language": None}
        assert label1 != label2

    def testUriEquality(self):
        label1 = Label(
            "Knokke-Heist", uri="urn:x-skosprovider:gemeenten:Knokke-Heist:nl-BE"
        )
        label2 = Label(
            "Knokke-Heist", uri="urn:x-skosprovider:gemeenten:Knokke-Heist:nl"
        )
        label3 = Label(
            "Cnocke-Heyst",
            type="altLabel",
            language="vls",
            uri="urn:x-skosprovider:gemeenten:Knokke-Heist:nl-BE",
        )
        assert label1 != label2
        assert label1 == label3

    def testUriDictEquality(self):
        label1 = Label(
            "Knokke-Heist", uri="urn:x-skosprovider:gemeenten:Knokke-Heist:und"
        )
        label2 = {
            "label": "Knokke-Heist",
            "type": "prefLabel",
            "language": "und",
            "uri": "urn:x-skosprovider:gemeenten:Knokke-Heist:und",
        }
        assert label1 == label2

    def testUriDictInequality(self):
        label1 = Label(
            "Knokke-Heist", uri="urn:x-skosprovider:gemeenten:Knokke-Heist:nl-BE"
        )
        label2 = {"label": "Knokke-Heist", "type": "prefLabel", "language": "und"}
        assert label1 != label2


class TestNote:

    def testConstructor(self):
        note = Note("Een gemeente in West-Vlaanderen.", type="note", language="nl-BE")
        assert "Een gemeente in West-Vlaanderen." == note.note
        assert "note" == note.type
        assert "nl-BE" == note.language

    def testConstructorInvalidLanguage(self):
        with pytest.raises(ValueError):
            Note("Een gemeente in West-Vlaanderen.", type="note", language="nederlands")
        note = Note("Een gemeente in West-Vlaanderen.", type="note", language=None)
        assert note.language == "und"

    def testConstructorInvalidMarkup(self):
        with pytest.raises(ValueError):
            Note(
                "Een gemeente in West-Vlaanderen.",
                type="note",
                language="nl",
                markup="markdown",
            )

    def testEquality(self):
        note1 = Note("A note.")
        note2 = Note("A note.", "note", "und")
        assert note1 == note2

    def testInEquality(self):
        note1 = Note("A note.")
        note2 = Note("A note.", "definition", "und")
        assert note1 != note2

    def testDictEquality(self):
        note1 = Note("A note.")
        note2 = {"note": "A note.", "type": "note", "language": "und", "markup": None}
        assert note1 == note2

    def testDictInequality(self):
        note1 = Note("A note.")
        note2 = {
            "note": "A note.",
            "type": "definition",
            "language": "und",
            "markup": None,
        }
        assert note1 != note2

    def testConstructorWithHTML(self):
        note = Note(
            "<p>Een gemeente in <em>West-Vlaanderen</em>.</p>",
            type="note",
            language="nl-BE",
            markup="HTML",
        )
        assert "<p>Een gemeente in <em>West-Vlaanderen</em>.</p>" == note.note
        assert "note" == note.type
        assert "nl-BE" == note.language
        assert "HTML" == note.markup

    def testIsValidType(self):
        assert Note.is_valid_type("note")
        assert not Note.is_valid_type("notitie")
        note = Note("A community in West-Flanders.", "definition", "en")
        assert note.is_valid_type("definition")

    def testIsValidMarkup(self):
        assert Note.is_valid_markup("HTML")
        assert not Note.is_valid_markup("markdown")
        note = Note("A community in West-Flanders.", "definition", "en", None)
        assert note.is_valid_markup(None)


class TestSource:

    def testConstructor(self):
        citation = "Van Daele, K; Meganck, L. & Mortier, S. 2015. "
        "Data-driven systems and system-driven data: the story of the "
        "Flanders Heritage Inventory (1995-2015)"
        source = Source(citation)
        assert citation == source.citation

    def testConstructorWithHTML(self):
        citation = "Van Daele, K; Meganck, L. & Mortier, S. 2015. "
        "<em>Data-driven systems and system-driven data: the story of the "
        "Flanders Heritage Inventory (1995-2015)</em>"
        source = Source(citation, markup="HTML")
        assert citation == source.citation
        assert "HTML" == source.markup

    def testIsValidMarkup(self):
        assert Source.is_valid_markup("HTML")
        assert not Source.is_valid_markup("markdown")
        citation = "Van Daele, K; Meganck, L. & Mortier, S. 2015. "
        "Data-driven systems and system-driven data: the story of the "
        "Flanders Heritage Inventory (1995-2015)"
        source = Source(citation)
        assert source.is_valid_markup(None)

    def testConstructorInvalidMarkup(self):
        with pytest.raises(ValueError):
            citation = "Van Daele, K; Meganck, L. & Mortier, S. 2015. "
            "<em>Data-driven systems and system-driven data: the story of the "
            "Flanders Heritage Inventory (1995-2015)</em>"
            Source(citation, markup="markdown")


class TestConceptScheme:

    def _get_gemeenten_nl(self):
        return Label("Gemeenten", type="prefLabel", language="nl-BE")

    def _get_fusiegemeenten_nl(self):
        return Label("Fusiegemeenten", type="altLabel", language="nl-BE")

    def _get_communities_en(self):
        return Label("Communities", type="prefLabel", language="en")

    def _get_labels(self):
        return [
            self._get_gemeenten_nl(),
            self._get_fusiegemeenten_nl(),
            self._get_communities_en(),
        ]

    def testRepr(self):
        conceptscheme = ConceptScheme(uri="urn:x-skosprovider:gemeenten")
        assert "ConceptScheme('urn:x-skosprovider:gemeenten')" == repr(conceptscheme)

    def testLabel(self):
        labels = self._get_labels()
        conceptscheme = ConceptScheme(uri="urn:x-skosprovider:gemeenten", labels=labels)
        assert label(labels) == conceptscheme.label()
        assert label(labels, "nl") == conceptscheme.label("nl")
        assert label(labels, "en") == conceptscheme.label("en")
        assert label(labels, None) == conceptscheme.label(None)

    def testSortKey(self):
        labels = self._get_labels()
        sortlabel = Label("allereerste", type="sortLabel", language="nl-BE")
        labels.append(sortlabel)
        conceptscheme = ConceptScheme(uri="urn:x-skosprovider:gemeenten", labels=labels)
        assert "allereerste" == conceptscheme._sortkey("sortlabel")
        assert "allereerste", conceptscheme._sortkey("sortlabel", "nl")
        assert "communities", conceptscheme._sortkey("sortlabel", "en")
        assert "urn:x-skosprovider:gemeenten", conceptscheme._sortkey("uri")

    def testLanguages(self):
        labels = self._get_labels()
        conceptscheme = ConceptScheme(
            uri="urn:x-skosprovider:gemeenten",
            labels=labels,
            languages=["nl", "en", "und"],
        )
        assert conceptscheme.languages == ["nl", "en", "und"]

    def testSource(self):
        conceptscheme = ConceptScheme(
            uri="urn:x-skosprovider:gemeenten", sources=[{"citation": "My citation"}]
        )
        assert 1 == len(conceptscheme.sources)
        assert "My citation" == conceptscheme.sources[0].citation

    def testEmptyUri(self):
        with pytest.raises(ValueError):
            ConceptScheme(uri=None)


class TestConcept:

    def _get_knokke_heist_nl(self):
        return Label("Knokke-Heist", type="prefLabel", language="nl-BE")

    def _get_cnocke_heyst_nl(self):
        return Label("Cnock-Heyst", type="altLabel", language="nl-BE")

    def _get_knokke_heist_en(self):
        return Label("Knocke-Heyst", type="prefLabel", language="en")

    def _get_labels(self):
        return [
            self._get_knokke_heist_nl(),
            self._get_cnocke_heyst_nl(),
            self._get_knokke_heist_en(),
        ]

    def testRepr(self):
        concept = Concept(1)
        assert "Concept('1')" == repr(concept)

    def testIn(self):
        c = Concept(1)
        assert hasattr(c, "id")
        assert hasattr(c, "uri")
        assert hasattr(c, "labels")
        assert hasattr(c, "notes")
        assert hasattr(c, "broader")
        assert hasattr(c, "narrower")
        assert hasattr(c, "related")
        assert hasattr(c, "member_of")

    def testLabel(self):
        labels = self._get_labels()
        c = Concept(1, labels=labels)
        assert label(labels) == c.label()
        assert label(labels, "nl") == c.label("nl")
        assert label(labels, "en") == c.label("en")
        assert label(labels, None) == c.label(None)

    def testSortKey(self):
        labels = self._get_labels()
        sl = Label("allereerste", type="sortLabel", language="nl-BE")
        labels.append(sl)
        c = Concept(1, labels=labels)
        assert "allereerste" == c._sortkey("sortlabel")
        assert "allereerste" == c._sortkey("sortlabel", "nl")
        assert "knocke-heyst" == c._sortkey("sortlabel", "en")
        assert "" == c._sortkey("uri")

    def testUri(self):
        c = Concept(1, uri="urn:x-skosprovider:gemeenten:1")
        assert 1 == c.id
        assert "urn:x-skosprovider:gemeenten:1" == c.uri

    def testMemberOf(self):
        c = Concept(1, uri="urn:x-skosprovider:gemeenten:1", member_of=[15])
        assert {15} == set(c.member_of)

    def testMatches(self):
        c = Concept(
            1,
            uri="urn:x-skosprovider:gemeenten:1",
            matches={"broad": ["http://id.something.org/provincies/1"]},
        )
        assert "close" in c.matches
        assert "exact" in c.matches
        assert "broad" in c.matches
        assert "narrow" in c.matches
        assert "related" in c.matches
        assert ["http://id.something.org/provincies/1"] == c.matches["broad"]

    def testSource(self):
        c = Concept(id=1, sources=[{"citation": "My citation"}])
        assert 1 == len(c.sources)
        assert "My citation" == c.sources[0].citation


class TestCollection:

    def _get_deelgemeenten_nl(self):
        return Label("Deelgemeenten", type="prefLabel", language="nl-BE")

    def _get_prefusiegemeenten_nl(self):
        return Label("Prefusiegemeenten", type="altLabel", language="nl-BE")

    def _get_labels(self):
        return [
            self._get_deelgemeenten_nl(),
            self._get_prefusiegemeenten_nl(),
        ]

    def testRepr(self):
        collection = Collection(1)
        assert "Collection('1')" == repr(collection)

    def testId(self):
        coll = Collection(350)
        assert 350 == coll.id

    def testUri(self):
        collection = Collection(350, uri="urn:x-skosprovider:gemeenten:350")
        assert 350 == collection.id
        assert "urn:x-skosprovider:gemeenten:350" == collection.uri

    def testLabel(self):
        labels = self._get_labels()
        coll = Collection(350, labels=labels)
        assert label(labels) == coll.label()
        assert label(labels, "nl") == coll.label("nl")
        assert label(labels, "en") == coll.label("en")
        assert label(labels, None) == coll.label(None)

    def testSortkey(self):
        labels = self._get_labels()
        sortlabel = Label("allereerste", type="sortLabel", language="nl-BE")
        labels.append(sortlabel)
        coll = Collection(350, labels=labels)
        assert "allereerste" == coll._sortkey("sortlabel")
        assert "allereerste" == coll._sortkey("sortlabel", "nl")
        assert "allereerste" == coll._sortkey("sortlabel", "en")
        assert "deelgemeenten" == coll._sortkey("label", "nl")
        assert "" == coll._sortkey("uri")

    def testEmptyMembers(self):
        labels = self._get_labels()
        coll = Collection(350, labels=labels, members=[])
        assert [] == coll.members

    def testMembers(self):
        labels = self._get_labels()
        coll = Collection(id=350, labels=labels, members=[1, 2])
        assert {1, 2} == set(coll.members)

    def testMemberOf(self):
        coll = Collection(id=1, member_of=[350])
        assert {350} == set(coll.member_of)

    def testSource(self):
        coll = Collection(id=1, sources=[{"citation": "My citation"}])
        assert 1 == len(coll.sources)
        assert "My citation" == coll.sources[0].citation

    def testnferConceptRelations(self):
        coll = Collection(
            id=1,
        )
        assert coll.infer_concept_relations
        coll = Collection(id=1, infer_concept_relations=False)
        assert not coll.infer_concept_relations


class TestDictToNoteFunction:

    def testDictToNodeWithDict(self):
        note = dict_to_note({"note": "A note.", "type": "note"})
        assert "A note." == note.note
        assert "note" == note.type
        assert "und" == note.language

    def testDictToNodeWithNote(self):
        note = dict_to_note(Note("A note.", "note"))
        assert "A note." == note.note
        assert "note" == note.type
        assert "und" == note.language


class TestDictToLabelFunction:

    def testDictToLabelWithDict(self):
        label = dict_to_label({"label": "A label.", "type": "prefLabel"})
        assert "A label." == label.label
        assert "prefLabel" == label.type
        assert "und" == label.language

    def testDictToLabelWithlabel(self):
        label = dict_to_label(Label("A label.", "prefLabel"))
        assert "A label." == label.label
        assert "prefLabel" == label.type
        assert "und" == label.language


class TestDictToSourceFunction:

    def testDictToSourceWithDict(self):
        citation = "Van Daele, K; Meganck, L. & Mortier, S. 2015. "
        "Data-driven systems and system-driven data: the story of the "
        "Flanders Heritage Inventory (1995-2015)"
        source = dict_to_source({"citation": citation})
        assert citation == source.citation

    def testDictToSourceWithDictWithMarkup(self):
        citation = "<strong>Van Daele, K; Meganck, L. & Mortier, S.</strong> 2015. "
        "Data-driven systems and system-driven data: the story of the "
        "Flanders Heritage Inventory (1995-2015)"
        source = dict_to_source({"citation": citation, "markup": "HTML"})
        assert citation == source.citation
        assert "HTML" == source.markup

    def testDictToSourceWithSource(self):
        citation = "Van Daele, K; Meganck, L. & Mortier, S. 2015. "
        "Data-driven systems and system-driven data: the story of the "
        "Flanders Heritage Inventory (1995-2015)"
        source = dict_to_source(Source(citation))
        assert citation == source.citation


class TestLabelFunction:

    def _get_knokke_heist_nl(self):
        return Label("Knokke-Heist", type="prefLabel", language="nl-BE")

    def _get_cnocke_heyst_nl(self):
        return Label("Cnock-Heyst", type="altLabel", language="nl-BE")

    def _get_knokke_heist_en(self):
        return Label("Knocke-Heyst", type="prefLabel", language="en-GB")

    def _get_und(self):
        return Label("nokke-eist", type="prefLabel", language="und")

    def _get_sortlabel(self):
        return Label("allereerste", type="sortLabel", language="nl-BE")

    def test_label_empty(self):
        assert label([]) is None
        assert label([], "nl-BE") is None
        assert label([], None) is None
        assert label([], "und") is None
        assert label([], ["nl-BE"]) is None

    def test_label_pref(self):
        kh = self._get_knokke_heist_nl()
        labels = [kh]
        assert kh == label(labels)
        assert kh == label(labels, "nl-BE")
        assert kh == label(labels, ["nl-BE"])
        assert kh == label(labels, "en-GB")
        assert kh == label(labels, ["en-GB"])
        assert kh == label(labels, None)

    def test_label_pref_und(self):
        und = self._get_und()
        labels = [und]
        assert label(labels) is not None
        assert und == label(labels)
        assert und == label(labels, "nl-BE")
        assert und == label(labels, ["nl-BE"])
        assert und == label(labels, "en-GB")
        assert und == label(labels, "und")
        assert und == label(labels, ["und"])
        assert und == label(labels, "any")
        assert und == label(labels, ["any"])
        assert und == label(labels, None)

    def test_label_pref_nl_and_en(self):
        kh = self._get_knokke_heist_nl()
        khen = self._get_knokke_heist_en()
        labels = [kh, khen]
        assert label(labels) in [kh, khen]
        assert kh == label(labels, "nl-BE")
        assert kh == label(labels, ["nl-BE", "en-GB"])
        assert khen == label(labels, "en-GB")
        assert khen == label(labels, ["fr", "en-GB"])
        assert label(labels, None) in [kh, khen]

    def test_label_inexact_language_match(self):
        kh = self._get_knokke_heist_nl()
        ch = self._get_cnocke_heyst_nl()
        khen = self._get_knokke_heist_en()
        labels = [kh, ch, khen]
        assert khen == label(labels, ["en", "nl"])
        assert khen == label(labels, "en")
        assert kh == label(labels, ["nl"])
        assert kh == label(labels, "nl")
        assert label(labels, None) in [kh, khen]

    def test_exact_precedes_inexact_match(self):
        khnl = Label("Knokke-Heist", type="prefLabel", language="nl")
        chnl = Label("Cnock-Heyst", type="altLabel", language="nl")
        khen = Label("Knocke-Heyst", type="prefLabel", language="en")
        khnlbe = self._get_knokke_heist_nl()
        chnlbe = self._get_cnocke_heyst_nl()
        khengb = self._get_knokke_heist_en()
        labels = [chnl, khen, khnlbe, khnl, chnlbe, khengb]
        assert khnlbe == label(labels, "nl-BE")
        assert khnlbe == label(labels, ["nl-BE"])
        assert khnl == label(labels, "nl")
        assert label(labels, "en-US") in [khen, khengb]

    def test_label_alt(self):
        ch = self._get_cnocke_heyst_nl()
        labels = [ch]
        assert ch == label(labels)
        assert ch == label(labels, "nl-BE")
        assert ch == label(labels, ["nl-BE"])
        assert ch == label(labels, "en-GB")
        assert ch == label(labels, ["en-GB"])
        assert ch == label(labels, None)

    def test_pref_precedes_alt(self):
        kh = self._get_knokke_heist_nl()
        ch = self._get_cnocke_heyst_nl()
        labels = [kh, ch]
        assert kh, label(labels)
        assert kh, label(labels, "nl-BE")
        assert kh, label(labels, ["nl-BE"])
        assert kh, label(labels, "en-GB")
        assert kh, label(labels, None)

    def test_sortlabel_unused(self):
        kh = self._get_knokke_heist_nl()
        ch = self._get_cnocke_heyst_nl()
        sl = self._get_sortlabel()
        labels = [kh, ch, sl]
        assert kh == label(labels)
        assert kh == label(labels, sortLabel=False)
        assert kh == label(labels, "nl-BE")
        assert kh == label(labels, ["nl-BE"])
        assert kh == label(labels, "en-GB")
        assert kh == label(labels, None)

    def test_sortlabel_broader(self):
        """
        Test that a broader sortlabel gets picked up for a regional sort.
        """
        kh = self._get_knokke_heist_nl()
        ch = self._get_cnocke_heyst_nl()
        sl = Label("Eerst", type="sortLabel", language="nl")
        labels = [kh, ch, sl]
        assert sl == label(labels, "nl-BE", True)
        assert sl == label(labels, ["nl-BE"], True)

    def test_dict_pref(self):
        kh = self._get_knokke_heist_nl()
        khd = kh.__dict__
        labels = [khd]
        assert kh == label(labels)
        assert kh == label(labels, "nl-BE")
        assert kh == label(labels, "en-GB")
        assert kh == label(labels, None)

    def test_find_best_label_for_type(self):
        kh = self._get_knokke_heist_nl()
        ch = self._get_cnocke_heyst_nl()
        khen = self._get_knokke_heist_en()
        labels = [kh, ch, khen]
        assert khen == find_best_label_for_type(labels, "en", "prefLabel")
        assert not find_best_label_for_type(labels, "en", "sortLabel")
        assert ch == find_best_label_for_type(labels, "nl", "altLabel")

    def test_filter_labels_by_language(self):
        kh = self._get_knokke_heist_nl()
        ch = self._get_cnocke_heyst_nl()
        khen = self._get_knokke_heist_en()
        labels = [kh, ch, khen]
        assert [kh, ch] == filter_labels_by_language(labels, "nl-BE")
        assert [] == filter_labels_by_language(labels, "nl")
        assert [kh, ch] == filter_labels_by_language(labels, "nl", True)
        assert labels == filter_labels_by_language(labels, "any")

    def test_filter_labels_by_language_unexisting(self):
        kh = self._get_knokke_heist_nl()
        ch = self._get_cnocke_heyst_nl()
        khen = self._get_knokke_heist_en()
        labels = [kh, ch, khen]
        assert label(labels, "tomatensoep") is not None
        assert label(labels, "") is not None



'''This module provides custom exceptions for skos providers.

.. versionadded:: 0.5.0
'''


class ProviderUnavailableException(Exception):
    '''
    This exception can be raised by a provider if it's unable to provide
    the thesaurus. This can occur when an underlying resource is unavailable
    (database connection, webservice, ...). The message should contain some
    more information about the problem.
    '''

    def __init__(self, message):
        '''
        :param message: More information about the exception.
        '''
        self.message = message

    def __repr__(self):
        return self.message



from skosprovider.exceptions import ProviderUnavailableException


class TestExceptions:

    def test_provider_unavailable_exception(self):
        exc = ProviderUnavailableException("test error")
        assert "test error" == repr(exc)



import unittest

from test_providers import geo
from test_providers import larch
from test_providers import trees


from skosprovider.providers import DictionaryProvider
from skosprovider.utils import add_lang_to_html
from skosprovider.utils import dict_dumper
from skosprovider.utils import extract_language


class DictDumperTest(unittest.TestCase):

    def setUp(self):
        self.larch_dump = {
            "id": "1",
            "uri": "http://id.trees.org/1",
            "type": "concept",
            "labels": [
                {"type": "prefLabel", "language": "en", "label": "The Larch"},
                {
                    "uri": "http://id.trees.org/labels/lariks-nl",
                    "type": "prefLabel",
                    "language": "nl",
                    "label": "De Lariks",
                },
            ],
            "notes": [
                {
                    "type": "definition",
                    "language": "en",
                    "note": "A type of tree.",
                    "markup": None,
                }
            ],
            "sources": [
                {
                    "citation": "Monthy Python. Episode Three: How to recognise "
                    "different types of trees from quite a long way away.",
                    "markup": None,
                }
            ],
            "narrower": [],
            "broader": [],
            "related": [],
            "member_of": ["3"],
            "subordinate_arrays": [],
            "matches": {
                "close": [],
                "exact": [
                    "http://id.python.org/different/types/of/trees/nr/1/the/larch"
                ],
                "related": [],
                "narrow": [],
                "broad": [],
            },
        }
        self.chestnut_dump = {
            "id": "2",
            "uri": "http://id.trees.org/2",
            "type": "concept",
            "labels": [
                {"type": "prefLabel", "language": "en", "label": "The Chestnut"},
                {"type": "altLabel", "language": "nl", "label": "De Paardekastanje"},
                {"type": "altLabel", "language": "fr", "label": "la châtaigne"},
            ],
            "notes": [
                {
                    "type": "definition",
                    "language": "en",
                    "note": "A different type of tree.",
                    "markup": None,
                }
            ],
            "sources": [
                {
                    "citation": '<span class="author">Bicycle repair man</span>',
                    "markup": "HTML",
                }
            ],
            "narrower": [],
            "broader": [],
            "related": [],
            "member_of": ["3"],
            "subordinate_arrays": [],
            "matches": {
                "close": [],
                "exact": [],
                "related": [
                    "http://id.python.org/different/types/of/trees/nr/17/the/other/chestnut"
                ],
                "narrow": [],
                "broad": [],
            },
        }
        self.species_dump = {
            "id": 3,
            "uri": "http://id.trees.org/3",
            "labels": [
                {"type": "prefLabel", "language": "en", "label": "Trees by species"},
                {"type": "prefLabel", "language": "nl", "label": "Bomen per soort"},
                {"type": "sortLabel", "language": "nl", "label": "aaa"},
            ],
            "type": "collection",
            "notes": [
                {
                    "type": "editorialNote",
                    "language": "en",
                    "note": "As seen in <em>How to Recognise Different Types "
                    "of Trees from Quite a Long Way Away</em>.",
                    "markup": "HTML",
                }
            ],
            "sources": [],
            "members": ["1", "2"],
            "member_of": [],
            "superordinates": [],
            "infer_concept_relations": False,
        }
        self.world_dump = {
            "id": "1",
            "uri": "urn:x-skosprovider:geography:1",
            "type": "concept",
            "labels": [{"type": "prefLabel", "language": "en", "label": "World"}],
            "notes": [],
            "sources": [],
            "narrower": [2, 3],
            "broader": [],
            "related": [],
            "member_of": [],
            "matches": {
                "close": [],
                "exact": [],
                "related": [],
                "narrow": [],
                "broad": [],
            },
            "subordinate_arrays": [],
        }

    def tearDown(self):
        del self.larch_dump
        del self.chestnut_dump
        del self.world_dump

    def _get_flat_provider(self, dictionary):
        return DictionaryProvider({"id": "TEST"}, dictionary)

    def _get_tree_provider(self, dictionary):
        return DictionaryProvider({"id": "TEST"}, dictionary)

    def testEmptyProvider(self):
        pv = self._get_flat_provider([])
        self.assertEqual([], dict_dumper(pv))

    def testOneElementProvider(self):
        pv = self._get_flat_provider([larch])
        assert [self.larch_dump] == dict_dumper(pv)

    def testFlatProvider(self):
        assert dict_dumper(trees) == [
            self.larch_dump,
            self.chestnut_dump,
            self.species_dump,
        ]

    def testEmptyTreeprovider(self):
        pv = self._get_tree_provider([])
        self.assertEqual([], dict_dumper(pv))

    def testTreeProvider(self):
        dump = dict_dumper(geo)
        self.assertIsInstance(dump, list)
        for c in dump:
            self.assertIsInstance(c, dict)
            self.assertIn("type", c)
            self.assertIn("id", c)
        self.assertIn(self.world_dump, dump)

    def testFlatProviderRoundTrip(self):
        dump = dict_dumper(trees)
        dump2 = dict_dumper(self._get_flat_provider(dict_dumper(trees)))
        self.assertEqual(dump, dump2)

    def testTreeProviderRoundTrip(self):
        dump = dict_dumper(geo)
        dump2 = dict_dumper(self._get_tree_provider(dict_dumper(geo)))
        self.assertEqual(dump, dump2)


class TestExtractLanguage:

    def test_extract_language_nlBE(self):
        assert "nl-BE" == extract_language("nl-BE")

    def test_extract_language_None(self):
        assert "und" == extract_language(None)


class TestHtml:

    def test_lang_und(self):
        assert "" == add_lang_to_html("", "und")
        assert "<p></p>" == add_lang_to_html("<p></p>", "und")

    def test_lang_no_html(self):
        assert '<div xml:lang="en"></div>' == add_lang_to_html("", "en")

    def test_no_single_child(self):
        html = "<p>Paragraph 1</p><p>Paragraph2</p>"
        assert (
            '<div xml:lang="en"><p>Paragraph 1</p><p>Paragraph2</p></div>'
            == add_lang_to_html(html, "en")
        )

    def test_text_node(self):
        html = "Something"
        assert '<div xml:lang="en">Something</div>' == add_lang_to_html(html, "en")

    def test_single_child_no_attributes(self):
        html = "<p>Paragraph 1</p>"
        assert '<p xml:lang="en">Paragraph 1</p>' == add_lang_to_html(html, "en")

    def test_single_child_already_has_langs(self):
        html = '<p xml:lang="en">Paragraph 1</p>'
        assert '<p xml:lang="en">Paragraph 1</p>' == add_lang_to_html(html, "en")

    def test_single_child_other_attributes(self):
        html = '<p class="something">Paragraph 1</p>'
        assert '<p class="something" xml:lang="en">Paragraph 1</p>' == add_lang_to_html(
            html, "en"
        )



'''
This example demonstrates the skosprovider API with a simple
DictionaryProvider containing just three items.
'''

from skosprovider.providers import DictionaryProvider
from skosprovider.uri import UriPatternGenerator
from skosprovider.skos import ConceptScheme

larch = {
    'id': '1',
    'uri': 'http://id.trees.org/1',
    'labels': [
        {'type': 'prefLabel', 'language': 'en', 'label': 'The Larch'},
        {'type': 'prefLabel', 'language': 'nl', 'label': 'De Lariks'}
    ],
    'notes': [
        {'type': 'definition', 'language': 'en', 'note': 'A type of tree.'}
    ],
    'member_of': ['3'],
    'matches': {
        'close': ['http://id.python.org/different/types/of/trees/nr/1/the/larch']
    }
}

chestnut = {
    'id': '2',
    'uri': 'http://id.trees.org/2',
    'labels': [
        {'type': 'prefLabel', 'language': 'en', 'label': 'The Chestnut'},
        {'type': 'altLabel', 'language': 'nl', 'label': 'De Paardekastanje'},
        {'type': 'altLabel', 'language': 'fr', 'label': 'la châtaigne'}
    ],
    'notes': [
        {
            'type': 'definition', 'language': 'en',
            'note': 'A different type of tree.'
        }
    ],
    'member_of': ['3'],
    'matches': {
        'related': ['http://id.python.org/different/types/of/trees/nr/17/the/other/chestnut']
    }
}

species = {
    'id': 3,
    'uri': 'http://id.trees.org/3',
    'labels': [
        {'type': 'prefLabel', 'language': 'en', 'label': 'Trees by species'},
        {'type': 'prefLabel', 'language': 'nl', 'label': 'Bomen per soort'}
    ],
    'type': 'collection',
    'members': ['1', '2'],
    'notes': [
        {
            'type': 'editorialNote',
            'language': 'en',
            'note': 'As seen in <em>How to Recognise Different Types of Trees from Quite a Long Way Away</em>.',
            'markup': 'HTML'
        }
    ]
}

provider = DictionaryProvider(
    {
        'id': 'TREES',
        'default_language': 'nl',
        'subject': ['biology']
    },
    [larch, chestnut, species],
    uri_generator=UriPatternGenerator('http://id.trees.org/types/%s'),
    concept_scheme=ConceptScheme('http://id.trees.org')
)

# Get a concept or collection by id
print(provider.get_by_id(1).label().label)

# Get a concept or collection by uri
print(provider.get_by_uri('http://id.trees.org/types/1'))

# Get all concepts and collections in a provider
# If possible, show a Dutch(as spoken in Belgium) label
# Order them ascending by label
print(provider.get_all(language='nl-BE', sort='label', sort_order='asc'))

# Get the top concepts in a provider
print(provider.get_top_concepts())

# Find anything that has a label of horse
print(provider.find({'label': 'The Larch'}))

# Get the top of a display hierarchy
print(provider.get_top_display())

# Get the children to display in a hierarchy
# If possible, show a French, Dutch or German label
# Order them descending by id
print(provider.get_children_display(3, language=['fr-BE', 'nl-BE', 'de-BE'], sort='id', sort_order='desc'))

# Get all concepts underneath a concept or collection
print(provider.expand(3))



import csv
import os
import unittest

from skosprovider.providers import DictionaryProvider
from skosprovider.providers import SimpleCsvProvider
from skosprovider.skos import Collection
from skosprovider.skos import Concept
from skosprovider.skos import ConceptScheme
from skosprovider.skos import Note

larch = {
    "id": "1",
    "uri": "http://id.trees.org/1",
    "labels": [
        {"type": "prefLabel", "language": "en", "label": "The Larch"},
        {
            "uri": "http://id.trees.org/labels/lariks-nl",
            "type": "prefLabel",
            "language": "nl",
            "label": "De Lariks",
        },
    ],
    "notes": [
        {
            "type": "definition",
            "language": "en",
            "note": "A type of tree.",
            "markup": None,
        }
    ],
    "sources": [
        {
            "citation": "Monthy Python. Episode Three: How to recognise different "
            "types of trees from quite a long way away."
        }
    ],
    "member_of": ["3"],
    "matches": {
        "exact": ["http://id.python.org/different/types/of/trees/nr/1/the/larch"]
    },
}

chestnut = {
    "id": "2",
    "uri": "http://id.trees.org/2",
    "labels": [
        {"type": "prefLabel", "language": "en", "label": "The Chestnut"},
        {"type": "altLabel", "language": "nl", "label": "De Paardekastanje"},
        {"type": "altLabel", "language": "fr", "label": "la châtaigne"},
    ],
    "notes": [
        {"type": "definition", "language": "en", "note": "A different type of tree."}
    ],
    "sources": [
        {"citation": '<span class="author">Bicycle repair man</span>', "markup": "HTML"}
    ],
    "member_of": ["3"],
    "matches": {
        "related": [
            "http://id.python.org/different/types/of/trees/nr/17/the/other/chestnut"
        ]
    },
}

species = {
    "id": 3,
    "uri": "http://id.trees.org/3",
    "labels": [
        {"type": "prefLabel", "language": "en", "label": "Trees by species"},
        {"type": "prefLabel", "language": "nl", "label": "Bomen per soort"},
        {"type": "sortLabel", "language": "nl", "label": "aaa"},
    ],
    "type": "collection",
    "members": ["1", "2"],
    "notes": [
        {
            "type": "editorialNote",
            "language": "en",
            "note": "As seen in <em>How to Recognise Different Types of Trees "
            "from Quite a Long Way Away</em>.",
            "markup": "HTML",
        }
    ],
    "infer_concept_relations": False,
}

trees = DictionaryProvider(
    {
        "id": "TREES",
        "default_language": "nl",
        "subject": ["biology"],
        "dataset": {"uri": "http://id.trees.org/dataset"},
    },
    [larch, chestnut, species],
    concept_scheme=ConceptScheme(
        "http://id.trees.org",
        labels=[
            {
                "uri": "http://id.trees.org/labels/soorten-nl",
                "type": "prefLabel",
                "language": "nl",
                "label": "Soorten",
            },
            {"type": "prefLabel", "language": "en", "label": "Species"},
        ],
        languages=["nl", "en"],
    ),
)

world = {
    "id": "1",
    "uri": None,
    "labels": [{"type": "prefLabel", "language": "en", "label": "World"}],
    "narrower": [2, 3],
}

geo = DictionaryProvider(
    {"id": "GEOGRAPHY"},
    [
        world,
        {
            "id": 2,
            "labels": [{"type": "prefLabel", "language": "en", "label": "Europe"}],
            "narrower": [4, 5],
            "broader": [1],
        },
        {
            "id": 3,
            "labels": [
                {"type": "prefLabel", "language": "en", "label": "North-America"}
            ],
            "narrower": [6],
            "broader": [1],
        },
        {
            "id": 4,
            "labels": [{"type": "prefLabel", "language": "en", "label": "Belgium"}],
            "broader": [2],
            "narrower": [16],
            "member_of": ["333"],
            "subordinate_arrays": ["358", "359"],
        },
        {
            "id": 5,
            "labels": [
                {"type": "prefLabel", "language": "en", "label": "United Kingdom"},
                {"type": "prefLabel", "language": "und", "label": "Brittannia"},
            ],
            "narrower": [10, 11, 12],
            "broader": [2],
        },
        {
            "id": 6,
            "labels": [
                {
                    "type": "prefLabel",
                    "language": "en",
                    "label": "United States of America",
                }
            ],
            "broader": [3],
        },
        {
            "id": 7,
            "labels": [{"type": "prefLabel", "language": "en", "label": "Flanders"}],
            "member_of": ["333", "358"],
        },
        {
            "id": 8,
            "labels": [{"type": "prefLabel", "language": "en", "label": "Brussels"}],
            "member_of": ["333", "358"],
        },
        {
            "id": 9,
            "labels": [{"type": "prefLabel", "language": "en", "label": "Wallonie"}],
            "member_of": [358],
        },
        {
            "id": 10,
            "labels": [{"type": "prefLabel", "language": "en", "label": "Scotland"}],
            "broader": [5],
        },
        {
            "id": 11,
            "labels": [{"type": "prefLabel", "language": "en", "label": "England"}],
            "broader": [5],
        },
        {
            "id": 12,
            "labels": [{"type": "prefLabel", "language": "en", "label": "Wales"}],
            "broader": [5],
        },
        {
            "id": 13,
            "labels": [{"type": "prefLabel", "language": "en", "label": "French"}],
            "member_of": [359],
        },
        {
            "id": 14,
            "labels": [{"type": "prefLabel", "language": "en", "label": "Dutch"}],
            "member_of": [359],
        },
        {
            "id": 15,
            "labels": [{"type": "prefLabel", "language": "en", "label": "German"}],
            "member_of": [359],
        },
        {
            "id": 16,
            "labels": [{"type": "prefLabel", "language": "en", "label": "The coast"}],
            "broader": [4],
        },
        {
            "id": "333",
            "type": "collection",
            "labels": [
                {
                    "type": "prefLabel",
                    "language": "en",
                    "label": "Places where dutch is spoken",
                }
            ],
            "members": ["4", "7", "8"],
            "infer_concept_relations": False,
        },
        {
            "id": "358",
            "type": "collection",
            "labels": [
                {"type": "prefLabel", "language": "en", "label": "Gewesten of Belgium"}
            ],
            "members": ["7", "8", "9"],
            "superordinates": ["4"],
            "infer_concept_relations": True,
        },
        {
            "id": 359,
            "type": "collection",
            "labels": [
                {"type": "prefLabel", "language": "en", "label": "Languages of Belgium"}
            ],
            "members": [13, 14, 15],
            "superordinates": [4],
            "infer_concept_relations": False,
        },
    ],
)


class TreesDictionaryProviderTests(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_get_vocabulary_id(self):
        self.assertEqual("TREES", trees.get_vocabulary_id())

    def test_get_vocabulary_uri(self):
        assert trees.get_vocabulary_uri() == trees.concept_scheme.uri

    def test_override_get_vocabulary_uri(self):
        trees = DictionaryProvider(
            {
                "id": "TREES",
                "uri": "http://id.trees.org",
                "default_language": "nl",
                "subject": ["biology"],
                "dataset": {"uri": "http://id.trees.org/dataset"},
            },
            [larch, chestnut, species],
            concept_scheme=ConceptScheme(
                "http://idtoo.trees.org",
                labels=[{"type": "prefLabel", "language": "nl", "label": "Soorten"}],
                languages=["nl", "en"],
            ),
        )
        assert "http://id.trees.org" == trees.get_vocabulary_uri()
        assert "http://idtoo.trees.org" == trees.concept_scheme.uri

    def test_override_get_vocabulary_uri_generates_cs(self):
        trees = DictionaryProvider(
            {
                "id": "TREES",
                "uri": "http://id.trees.org",
                "default_language": "nl",
                "subject": ["biology"],
                "dataset": {"uri": "http://id.trees.org/dataset"},
            },
            [larch, chestnut, species],
        )
        assert "http://id.trees.org" == trees.get_vocabulary_uri()
        assert "http://id.trees.org" == trees.concept_scheme.uri

    def test_get_metadata(self):
        assert trees.get_metadata() == {
            "id": "TREES",
            "default_language": "nl",
            "subject": ["biology"],
            "dataset": {"uri": "http://id.trees.org/dataset"},
        }

    def test_allowed_instance_scopes(self):
        assert trees.allowed_instance_scopes == [
            "single",
            "threaded_thread",
            "threaded_global",
        ]

    def test_override_instance_scopes(self):
        trees = DictionaryProvider(
            {"id": "TREES"}, [larch], allowed_instance_scopes=["single"]
        )
        assert trees.allowed_instance_scopes == ["single"]

    def test_get_by_id(self):
        lariks = trees.get_by_id(1)
        assert larch["id"] == lariks.id
        assert larch["uri"] == lariks.uri
        assert larch["labels"] == lariks.labels
        assert larch["notes"] == lariks.notes
        assert len(larch["sources"]) == len(lariks.sources)
        assert larch["sources"][0]["citation"] == lariks.sources[0].citation

    def test_concept_has_scheme(self):
        lariks = trees.get_by_id(1)
        self.assertIsInstance(lariks.concept_scheme, ConceptScheme)
        self.assertEqual("http://id.trees.org", lariks.concept_scheme.uri)

    def test_collection_has_scheme(self):
        coll = trees.get_by_id(3)
        self.assertIsInstance(coll.concept_scheme, ConceptScheme)
        self.assertEqual("http://id.trees.org", coll.concept_scheme.uri)

    def test_get_by_uri(self):
        lariks = trees.get_by_uri("http://id.trees.org/1")
        assert "http://id.trees.org/1" == lariks.uri

    def test_get_by_id_string(self):
        lariks = trees.get_by_id("1")
        self.assertEqual(larch["id"], lariks.id)
        self.assertEqual(larch["uri"], lariks.uri)
        self.assertEqual(larch["labels"], lariks.labels)
        self.assertEqual(larch["notes"], lariks.notes)
        self.assertEqual(larch["member_of"], lariks.member_of)
        self.assertEqual("concept", lariks.type)
        assert 5 == len(lariks.matches)
        assert 1 == len(lariks.matches["exact"])
        assert larch["matches"]["exact"] == lariks.matches["exact"]
        assert [] == lariks.matches["related"]

    def test_get_by_id_is_type_agnostic(self):
        self.assertEqual(trees.get_by_id(1), trees.get_by_id("1"))

    def test_get_unexisting_by_id(self):
        self.assertFalse(trees.get_by_id(987654321))

    def test_get_unexisting_by_uri(self):
        self.assertFalse(trees.get_by_uri("urn:x-skosprovider:987654321"))

    def test_expand_concept(self):
        self.assertEqual(["1"], trees.expand(1))

    def test_expand_unexisting(self):
        self.assertEqual(False, trees.expand(987654321))

    def test_expand_collection(self):
        self.assertEqual({"1", "2"}, set(trees.expand(3)))

    def test_get_all(self):
        self.assertEqual(
            trees.get_all(),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                },
                {
                    "id": "2",
                    "uri": "http://id.trees.org/2",
                    "type": "concept",
                    "label": "De Paardekastanje",
                },
                {
                    "id": 3,
                    "uri": "http://id.trees.org/3",
                    "type": "collection",
                    "label": "Bomen per soort",
                },
            ],
        )

    def test_get_all_default_language(self):
        trees = DictionaryProvider({"id": "TREES"}, [larch])
        self.assertEqual(
            trees.get_all(),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "The Larch",
                }
            ],
        )

    def test_get_all_sort_label(self):
        self.assertEqual(
            trees.get_all(sort="label"),
            [
                {
                    "id": 3,
                    "uri": "http://id.trees.org/3",
                    "type": "collection",
                    "label": "Bomen per soort",
                },
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                },
                {
                    "id": "2",
                    "uri": "http://id.trees.org/2",
                    "type": "concept",
                    "label": "De Paardekastanje",
                },
            ],
        )

    def test_get_all_sort_id(self):
        self.assertEqual(
            trees.get_all(sort="id", sort_order="asc"),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                },
                {
                    "id": "2",
                    "uri": "http://id.trees.org/2",
                    "type": "concept",
                    "label": "De Paardekastanje",
                },
                {
                    "id": 3,
                    "uri": "http://id.trees.org/3",
                    "type": "collection",
                    "label": "Bomen per soort",
                },
            ],
        )

    def test_get_all_sort_id_reverse(self):
        self.assertEqual(
            trees.get_all(sort="id", sort_order="desc"),
            [
                {
                    "id": 3,
                    "uri": "http://id.trees.org/3",
                    "type": "collection",
                    "label": "Bomen per soort",
                },
                {
                    "id": "2",
                    "uri": "http://id.trees.org/2",
                    "type": "concept",
                    "label": "De Paardekastanje",
                },
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                },
            ],
        )

    def test_get_top_concepts_default_language(self):
        self.assertEqual(
            trees.get_top_concepts(),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                },
                {
                    "id": "2",
                    "uri": "http://id.trees.org/2",
                    "type": "concept",
                    "label": "De Paardekastanje",
                },
            ],
        )

    def test_get_top_concepts_sorted_by_id(self):
        self.assertEqual(
            trees.get_top_concepts(sort="id"),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                },
                {
                    "id": "2",
                    "uri": "http://id.trees.org/2",
                    "type": "concept",
                    "label": "De Paardekastanje",
                },
            ],
        )

    def test_get_all_english(self):
        self.assertEqual(
            trees.get_all(language="en"),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "The Larch",
                },
                {
                    "id": "2",
                    "uri": "http://id.trees.org/2",
                    "type": "concept",
                    "label": "The Chestnut",
                },
                {
                    "id": 3,
                    "uri": "http://id.trees.org/3",
                    "type": "collection",
                    "label": "Trees by species",
                },
            ],
        )

    def test_get_all_english_sorted_by_label(self):
        self.assertEqual(
            trees.get_all(language="en", sort="label"),
            [
                {
                    "id": "2",
                    "uri": "http://id.trees.org/2",
                    "type": "concept",
                    "label": "The Chestnut",
                },
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "The Larch",
                },
                {
                    "id": 3,
                    "uri": "http://id.trees.org/3",
                    "type": "collection",
                    "label": "Trees by species",
                },
            ],
        )

    def test_get_all_french(self):
        la_chataigne = {
            "id": "2",
            "uri": "http://id.trees.org/2",
            "type": "concept",
            "label": "la châtaigne",
        }
        assert la_chataigne in trees.get_all(language="fr")

    def test_find_all(self):
        c = trees.find({"type": "all"})
        self.assertEqual(3, len(c))

    def test_find_all_sort(self):
        c = trees.find({"type": "all"}, sort="id", sort_order="desc")
        self.assertEqual([3, "2", "1"], [cc["id"] for cc in c])
        c = trees.find({"type": "all"}, sort="sortlabel", sort_order="asc")
        self.assertEqual([3, "1", "2"], [cc["id"] for cc in c])
        c = trees.find({"type": "all"}, sort="sortlabel", sort_order="desc")
        self.assertEqual(["2", "1", 3], [cc["id"] for cc in c])

    def test_find_concepts(self):
        c = trees.find({"type": "concept"})
        self.assertEqual(2, len(c))

    def test_find_collections(self):
        c = trees.find({"type": "collection"})
        self.assertEqual(1, len(c))

    def test_find_type_None(self):
        c = trees.find({"type": None})
        assert len(c) == 3

    def test_find_larch(self):
        self.assertEqual(
            trees.find({"label": "The Larch"}),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                }
            ],
        )

    def test_find_The_Lar(self):
        self.assertEqual(
            trees.find({"label": "The Lar"}),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                }
            ],
        )

    def test_find_case_sensitive(self):
        trees = DictionaryProvider(
            {"id": "TREES", "default_language": "nl"},
            [larch, chestnut, species],
            case_insensitive=False,
        )
        self.assertEqual(
            trees.find({"label": "The Lar"}),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                }
            ],
        )
        self.assertEqual(trees.find({"label": "lar"}), [])

    def test_find_kastanje(self):
        trees = DictionaryProvider(
            {"id": "TREES", "default_language": "nl"}, [larch, chestnut, species]
        )
        concepts = trees.find({"label": "De Paardekastanje"})
        assert len(concepts) == 1

    def test_find_empty_label(self):
        c = trees.find({"label": ""})
        self.assertEqual(3, len(c))

    def test_find_lar(self):
        self.assertEqual(
            trees.find({"label": "lar"}),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                }
            ],
        )

    def test_find_es(self):
        c = trees.find({"label": "es"})
        self.assertEqual(2, len(c))

    def test_find_all_es(self):
        c = trees.find({"label": "es", "type": "all"})
        self.assertEqual(2, len(c))

    def test_find_concepts_es(self):
        c = trees.find({"label": "es", "type": "concept"})
        self.assertEqual(1, len(c))
        for cc in c:
            self.assertIsInstance(trees.get_by_id(cc["id"]), Concept)

    def test_find_collections_es(self):
        c = trees.find({"label": "es", "type": "collection"})
        self.assertEqual(1, len(c))
        for cc in c:
            self.assertIsInstance(trees.get_by_id(cc["id"]), Collection)

    def test_find_no_arguments(self):
        self.assertEqual(trees.find({}), trees.get_all())

    def test_find_in_collection(self):
        c = trees.find({"collection": {"id": 3}})
        self.assertEqual(2, len(c))
        for cc in c:
            self.assertIsInstance(trees.get_by_id(cc["id"]), Concept)

    def test_find_in_collection_es(self):
        c = trees.find({"collection": {"id": 3}, "label": "es"})
        self.assertEqual(1, len(c))
        for cc in c:
            self.assertIsInstance(trees.get_by_id(cc["id"]), Concept)

    def test_find_in_unexisting_collection(self):
        self.assertRaises(ValueError, trees.find, {"collection": {"id": 404}})

    def test_find_matches_without_uri(self):
        self.assertRaises(ValueError, trees.find, {"matches": {"type": "close"}})

    def test_find_matches_uri_not_present(self):
        concepts = trees.find(
            {"matches": {"uri": "https://id.erfgoed.net/thesauri/soorten/1"}}
        )
        assert 0 == len(concepts)

    def test_find_matches_uri_present(self):
        concepts = trees.find(
            {
                "matches": {
                    "uri": "http://id.python.org/different/"
                    "types/of/trees/nr/1/the/larch"
                }
            }
        )
        assert 1 == len(concepts)

    def test_find_matches_uri_and_type_present(self):
        concepts = trees.find(
            {
                "matches": {
                    "uri": "http://id.python.org/different/"
                    "types/of/trees/nr/1/the/larch",
                    "type": "exact",
                }
            }
        )
        assert 1 == len(concepts)

    def test_find_matches_uri_and_type_inheritance_present(self):
        c = trees.find(
            {
                "matches": {
                    "uri": "http://id.python.org/different/"
                    "types/of/trees/nr/1/the/larch",
                    "type": "close",
                }
            }
        )
        self.assertEqual(1, len(c))

    def test_find_matches_uri_and_wrong_type(self):
        c = trees.find(
            {
                "matches": {
                    "uri": "http://id.python.org/different/"
                    "types/of/trees/nr/1/the/larch",
                    "type": "related",
                }
            }
        )
        self.assertEqual(0, len(c))

    def test_get_display_top(self):
        top = trees.get_top_display()
        self.assertEqual(1, len(top))
        self.assertIn(
            {
                "id": 3,
                "type": "collection",
                "label": "Bomen per soort",
                "uri": "http://id.trees.org/3",
            },
            top,
        )

    def test_get_display_top_sorted_label(self):
        top = trees.get_top_display(sort="label", language="nl")
        self.assertEqual(1, len(top))
        self.assertIn(
            {
                "id": 3,
                "type": "collection",
                "label": "Bomen per soort",
                "uri": "http://id.trees.org/3",
            },
            top,
        )

    def test_get_display_children_unexisting_concept(self):
        self.assertFalse(trees.get_children_display(404))

    def test_get_display_children_concept(self):
        self.assertEqual([], trees.get_children_display(1))
        self.assertEqual([], trees.get_children_display(2))

    def test_get_display_children_collection(self):
        self.assertEqual(
            trees.get_children_display(3),
            [
                {
                    "id": "1",
                    "uri": "http://id.trees.org/1",
                    "type": "concept",
                    "label": "De Lariks",
                },
                {
                    "id": "2",
                    "uri": "http://id.trees.org/2",
                    "type": "concept",
                    "label": "De Paardekastanje",
                },
            ],
        )

    def test_get_display_children_collection_sort_custom(self):
        assert trees.get_children_display(
            3, language="nl", sort="sortlabel", sort_order="desc"
        ) == [
            {
                "id": "2",
                "uri": "http://id.trees.org/2",
                "type": "concept",
                "label": "De Paardekastanje",
            },
            {
                "id": "1",
                "uri": "http://id.trees.org/1",
                "type": "concept",
                "label": "De Lariks",
            },
        ]


class GeoDictionaryProviderTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_get_vocabulary_id(self):
        self.assertEqual("GEOGRAPHY", geo.get_vocabulary_id())

    def test_get_metadata(self):
        self.assertEqual({"id": "GEOGRAPHY", "subject": []}, geo.get_metadata())

    def test_concept_has_scheme(self):
        con = geo.get_by_id(1)
        self.assertIsInstance(con.concept_scheme, ConceptScheme)
        self.assertEqual("urn:x-skosprovider:geography", con.concept_scheme.uri)

    def test_collection_has_scheme(self):
        coll = geo.get_by_id(333)
        self.assertIsInstance(coll.concept_scheme, ConceptScheme)
        self.assertEqual("urn:x-skosprovider:geography", coll.concept_scheme.uri)

    def test_get_top_concepts(self):
        top = geo.get_top_concepts()
        self.assertEqual(4, len(top))
        self.assertIn(
            {
                "id": "1",
                "uri": "urn:x-skosprovider:geography:1",
                "type": "concept",
                "label": "World",
            },
            top,
        )
        self.assertIn(
            {
                "id": 15,
                "uri": "urn:x-skosprovider:geography:15",
                "type": "concept",
                "label": "German",
            },
            top,
        )

    def test_get_by_id(self):
        wereld = geo.get_by_id(1)
        self.assertEqual(world["id"], wereld.id)
        self.assertIsNotNone(wereld.uri)
        self.assertEqual(world["labels"], wereld.labels)
        self.assertEqual(world["narrower"], wereld.narrower)

    def test_get_belgium_by_id(self):
        belgium = geo.get_by_id(4)
        self.assertEqual(4, belgium.id)
        self.assertEqual({"333"}, set(belgium.member_of))

    def test_get_by_uri(self):
        wereld = geo.get_by_uri("urn:x-skosprovider:geography:1")
        self.assertEqual(world["id"], wereld.id)
        self.assertEqual(world["labels"], wereld.labels)
        self.assertEqual(world["narrower"], wereld.narrower)

    def test_get_collection_by_id(self):
        dutch_speaking = geo.get_by_id(333)
        self.assertEqual("333", dutch_speaking.id)
        self.assertEqual(["4", "7", "8"], dutch_speaking.members)

    def test_get_collection_by_uri(self):
        dutch_speaking = geo.get_by_uri("urn:x-skosprovider:geography:333")
        self.assertEqual("333", dutch_speaking.id)
        self.assertEqual(["4", "7", "8"], dutch_speaking.members)

    def test_expand_Belgium(self):
        self.assertEqual({4, 7, 8, 9, 16}, set(geo.expand(4)))

    def test_expand_UK(self):
        self.assertEqual({5, 10, 11, 12}, set(geo.expand(5)))

    def test_expand_string(self):
        self.assertEqual({4, 7, 8, 9, 16}, set(geo.expand("4")))

    def test_expand_unexisting(self):
        self.assertEqual(False, geo.expand(987654321))

    def test_expand_collection(self):
        self.assertEqual({4, 7, 8, 9, 16}, set(geo.expand(333)))

    def test_find_in_collection(self):
        c = geo.find({"collection": {"id": 333}})
        self.assertEqual(3, len(c))
        for cc in c:
            self.assertIsInstance(geo.get_by_id(cc["id"]), Concept)

    def test_find_in_collection_depth_all(self):
        c = geo.find({"collection": {"id": 333, "depth": "all"}})
        self.assertEqual(5, len(c))
        for cc in c:
            self.assertIsInstance(geo.get_by_id(cc["id"]), Concept)

    def test_find_in_collection_depth_all_wallon(self):
        c = geo.find({"collection": {"id": "333", "depth": "all"}, "label": "Wallon"})
        self.assertEqual(1, len(c))
        for cc in c:
            self.assertIsInstance(geo.get_by_id(cc["id"]), Concept)

    def test_get_display_top(self):
        top = geo.get_top_display()
        self.assertEqual(2, len(top))
        self.assertIn(
            {
                "id": "1",
                "uri": "urn:x-skosprovider:geography:1",
                "type": "concept",
                "label": "World",
            },
            top,
        )

    def test_get_display_children_unexisting_concept(self):
        self.assertFalse(geo.get_children_display(404))

    def test_get_display_children_concept(self):
        self.assertEqual(
            geo.get_children_display(1),
            [
                {
                    "id": 2,
                    "uri": "urn:x-skosprovider:geography:2",
                    "type": "concept",
                    "label": "Europe",
                },
                {
                    "id": 3,
                    "type": "concept",
                    "uri": "urn:x-skosprovider:geography:3",
                    "label": "North-America",
                },
            ],
        )

    def test_get_display_children_collection(self):
        self.assertEqual(
            [
                {
                    "id": 4,
                    "uri": "urn:x-skosprovider:geography:4",
                    "type": "concept",
                    "label": "Belgium",
                },
                {
                    "id": 7,
                    "uri": "urn:x-skosprovider:geography:7",
                    "type": "concept",
                    "label": "Flanders",
                },
                {
                    "id": 8,
                    "uri": "urn:x-skosprovider:geography:8",
                    "type": "concept",
                    "label": "Brussels",
                },
            ],
            geo.get_children_display(333),
        )

    def test_get_display_children_concept_with_thesaurus_array(self):
        children = geo.get_children_display(4)
        self.assertEqual(3, len(children))
        self.assertIn(
            {
                "id": "358",
                "uri": "urn:x-skosprovider:geography:358",
                "type": "collection",
                "label": "Gewesten of Belgium",
            },
            children,
        )
        self.assertIn(
            {
                "id": 359,
                "uri": "urn:x-skosprovider:geography:359",
                "type": "collection",
                "label": "Languages of Belgium",
            },
            children,
        )
        self.assertIn(
            {
                "id": 16,
                "uri": "urn:x-skosprovider:geography:16",
                "type": "concept",
                "label": "The coast",
            },
            children,
        )


class SimpleCsvProviderTests(unittest.TestCase):

    def setUp(self):
        from skosprovider.uri import UriPatternGenerator

        self.ifile = open(os.path.join(os.path.dirname(__file__), "data", "menu.csv"))
        reader = csv.reader(self.ifile)
        self.csvprovider = SimpleCsvProvider(
            {"id": "MENU"},
            reader,
            uri_generator=UriPatternGenerator("http://id.python.org/menu/%s"),
            concept_scheme=ConceptScheme("http://id.python.org/menu"),
        )

    def tearDown(self):
        self.ifile.close()
        del self.csvprovider

    def testCount(self):
        self.assertEqual(11, len(self.csvprovider.get_all()))

    def testGetEggAndBacon(self):
        eb = self.csvprovider.get_by_id(1)
        self.assertIsInstance(eb, Concept)
        self.assertEqual("1", eb.id)
        self.assertEqual("http://id.python.org/menu/1", eb.uri)
        self.assertEqual("Egg and Bacon", eb.label().label)
        self.assertEqual("prefLabel", eb.label().type)
        self.assertEqual([], eb.notes)
        assert 1 == len(eb.sources)
        assert "Monthy Python, Episode Twenty-five." == eb.sources[0].citation

    def testGetEggAndSpamByUri(self):
        eb = self.csvprovider.get_by_uri("http://id.python.org/menu/3")
        self.assertIsInstance(eb, Concept)
        self.assertEqual("3", eb.id)
        self.assertEqual("http://id.python.org/menu/3", eb.uri)

    def testFindSpam(self):
        spam = self.csvprovider.find({"label": "Spam"})
        self.assertEqual(8, len(spam))

    def testGetLobster(self):
        eb = self.csvprovider.get_by_id(11)
        self.assertIsInstance(eb, Concept)
        self.assertEqual("11", eb.id)
        self.assertEqual("Lobster Thermidor", eb.label().label)
        self.assertIsInstance(eb.notes[0], Note)
        self.assertIn("Mornay", eb.notes[0].note)
        self.assertEqual("note", eb.notes[0].type)

    def testFindSausageCaseInsensitive(self):
        sausages = self.csvprovider.find({"label": "sausage"})
        self.assertEqual(4, len(sausages))

    def testFindSausageCaseSensitive(self):
        self.csvprovider.case_insensitive = False
        sausages = self.csvprovider.find({"label": "Sausage"})
        self.assertEqual(1, len(sausages))



"""
This module contains a read-only model of the :term:`SKOS` specification.

To complement the :term:`SKOS` specification, some elements were borrowed
from the :term:`SKOS-THES` specification (eg. superordinate and
subordinate array).

.. versionadded:: 0.2.0
"""

from language_tags import tags

from .uri import is_uri

valid_markup = [None, "HTML"]
"""
Valid types of markup for a note or a source.
"""


class Label:
    """
    A :term:`SKOS` Label.
    """

    uri = None
    """A :term:`URI` for this label."""

    label = None
    """
    The label itself (eg. `churches`, `trees`, `Spitfires`, ...)
    """

    type = "prefLabel"
    """
    The type of this label (`prefLabel`, `altLabel`, `hiddenLabel`, 'sortLabel').
    """

    language = "und"
    """
    The language the label is in (eg. `en`, `en-US`, `nl`, `nl-BE`).
    """

    valid_types = ["prefLabel", "altLabel", "hiddenLabel", "sortLabel"]
    """
    The valid types for a label
    """

    def __init__(self, label, type="prefLabel", language="und", uri=None):
        self.label = label
        self.type = type
        if not language:
            language = "und"
        if uri and not is_uri(uri):
            raise ValueError(f"{uri} is not a valid URI.")
        self.uri = uri
        if tags.check(language):
            self.language = language
        else:
            raise ValueError(f"{language} is not a valid IANA language tag.")

    def __eq__(self, other):
        if isinstance(other, dict):
            if self.uri:
                return self.uri == other.get("uri")
            return (
                self.label == other["label"]
                and self.type == other["type"]
                and self.language == other["language"]
            )
        if self.uri:
            return self.uri == other.uri
        return (
            self.label == other.label
            and self.type == other.type
            and self.language == other.language
        )

    def __ne__(self, other):
        return not self == other

    @staticmethod
    def is_valid_type(type):
        """
        Check if the argument is a valid SKOS label type.

        :param string type: The type to be checked.
        """
        return type in Label.valid_types

    def is_xl(self):
        return self.uri is not None

    def __repr__(self):
        if not self.is_xl():
            return f"Label('{self.label}', '{self.type}', '{self.language}')"
        return f"Label('{self.label}', '{self.type}', '{self.language}', '{self.uri}')"


class Note:
    """
    A :term:`SKOS` Note.
    """

    note = None
    """The note itself"""

    type = "note"
    """
    The type of this note ( `note`, `definition`, `scopeNote`, ...).
    """

    language = "und"
    """
    The language the label is in (eg. `en`, `en-US`, `nl`, `nl-BE`).
    """

    markup = None
    """
    What kind of markup does the note contain?

    If not `None`, the note should be treated as a certain type of markup.
    Currently only HTML is allowed.
    """

    valid_types = [
        "note",
        "changeNote",
        "definition",
        "editorialNote",
        "example",
        "historyNote",
        "scopeNote",
    ]
    """
    The valid types for a note.
    """

    def __init__(self, note, type="note", language="und", markup=None):
        self.note = note
        self.type = type
        if not language:
            language = "und"
        if tags.check(language):
            self.language = language
        else:
            raise ValueError(f"{language} is not a valid IANA language tag.")
        if self.is_valid_markup(markup):
            self.markup = markup
        else:
            raise ValueError(f"{markup} is not valid markup.")

    def __eq__(self, other):
        if isinstance(other, dict):
            return (
                self.note == other["note"]
                and self.type == other["type"]
                and self.language == other["language"]
            )
        return (
            self.note == other.note
            and self.type == other.type
            and self.language == other.language
        )

    def __ne__(self, other):
        return not self == other

    @staticmethod
    def is_valid_type(type):
        """
        Check if the argument is a valid SKOS note type.

        :param string type: The type to be checked.
        """
        return type in Note.valid_types

    @staticmethod
    def is_valid_markup(markup):
        """
        Check the argument is a valid type of markup.

        :param string markup: The type to be checked.
        """
        return markup in valid_markup


class Source:
    """
    A `Source` for a concept, collection or scheme.

    """

    citation = None
    """A bibliographic citation for this source."""

    markup = None
    """
    What kind of markup does the source contain?

    If not `None`, the source should be treated as a certain type of markup.
    Currently only HTML is allowed.
    """

    def __init__(self, citation, markup=None):
        self.citation = citation
        if self.is_valid_markup(markup):
            self.markup = markup
        else:
            raise ValueError(f"{markup} is not valid markup.")

    @staticmethod
    def is_valid_markup(markup):
        """
        Check the argument is a valid type of markup.

        :param string markup: The type to be checked.
        """
        return markup in valid_markup


class ConceptScheme:
    """
    A :term:`SKOS` ConceptScheme.

    :param string uri: A :term:`URI` for this conceptscheme.
    :param list labels: A list of :class:`skosprovider.skos.Label` instances.
    :param list notes: A list of :class:`skosprovider.skos.Note` instances.
    """

    uri = None
    """A :term:`URI` for this conceptscheme."""

    labels = []
    """A :class:`lst` of :class:`skosprovider.skos.label` instances."""

    notes = []
    """A :class:`lst` of :class:`skosprovider.skos.Note` instances."""

    sources = []
    """A :class:`lst` of :class:`skosprovider.skos.Source` instances."""

    languages = []
    """
    A :class:`lst` of languages that are being used in the ConceptScheme.

    There's no guarantuee that labels or notes in other languages do not exist.
    """

    def __init__(self, uri, labels=None, notes=None, sources=None, languages=None):
        if not is_uri(uri):
            raise ValueError(f"{uri} is not a valid URI.")
        self.uri = uri
        self.labels = [dict_to_label(label) for label in labels] if labels else []
        self.notes = [dict_to_note(note) for note in notes] if notes else []
        self.sources = [dict_to_source(source) for source in sources] if sources else []
        self.languages = languages or []

    def label(self, language="any"):
        """
        Provide a single label for this conceptscheme.

        This uses the :func:`label` function to determine which label to
        return.

        :param string language: The preferred language to receive the label in.
            This should be a valid IANA language tag.
        :rtype: :class:`skosprovider.skos.Label` or None if no labels were found.
        """
        return label(self.labels, language)

    def _sortkey(self, key="uri", language="any"):
        """
        Provide a single sortkey for this conceptscheme.

        :param string key: Either `uri`, `label` or `sortlabel`.
        :param string language: The preferred language to receive the label in
            if key is `label` or `sortlabel`. This should be a valid IANA language tag.
        :rtype: :class:`str`
        """
        if key == "uri":
            return self.uri
        else:
            sortlabel = label(self.labels, language, key == "sortlabel")
            return sortlabel.label.lower() if sortlabel else ""

    def __repr__(self):
        return f"ConceptScheme('{self.uri}')"


class Concept:
    """
    A :term:`SKOS` Concept.
    """

    id = None
    """An id for this Concept within a vocabulary

    eg. 12345
    """

    uri = None
    """A proper uri for this Concept

    eg. `http://id.example.com/skos/trees/1`
    """

    type = "concept"
    """The type of this concept or collection.

    eg. 'concept'
    """

    concept_scheme = None
    """The :class:`ConceptScheme` this Concept is a part of."""

    labels = []
    """A :class:`lst` of :class:`Label` instances."""

    notes = []
    """A :class:`lst` of :class:`Note` instances."""

    sources = []
    """A :class:`lst` of :class:`skosprovider.skos.Source` instances."""

    broader = []
    """A :class:`lst` of concept ids."""

    narrower = []
    """A :class:`lst` of concept ids."""

    related = []
    """A :class:`lst` of concept ids."""

    member_of = []
    """A :class:`lst` of collection ids."""

    subordinate_arrays = []
    """A :class:`list` of collection ids."""

    matches = ({},)
    """
    A :class:`dictionary`. Each key is a matchtype and
    contains a :class:`list` of URI's.
    """

    matchtypes = ["close", "exact", "related", "broad", "narrow"]
    """Matches with Concepts in other ConceptSchemes.

    This dictionary contains a key for each type of Match (close, exact,
    related, broad, narrow). Attached to each key is a list of URI's.
    """

    def __init__(
        self,
        id,
        uri=None,
        concept_scheme=None,
        labels=None,
        notes=None,
        sources=None,
        broader=None,
        narrower=None,
        related=None,
        member_of=None,
        subordinate_arrays=None,
        matches=None,
    ):
        self.id = id
        self.uri = uri
        self.type = "concept"
        self.concept_scheme = concept_scheme
        self.labels = [dict_to_label(label) for label in labels] if labels else []
        self.notes = [dict_to_note(note) for note in notes] if notes else []
        self.sources = [dict_to_source(source) for source in sources] if sources else []
        self.broader = broader or []
        self.narrower = narrower or []
        self.related = related or []
        self.member_of = member_of or []
        self.subordinate_arrays = subordinate_arrays or []
        self.matches = {key: [] for key in self.matchtypes}
        if matches:
            self.matches.update(matches)

    def label(self, language="any"):
        """
        Provide a single label for this concept.

        This uses the :func:`label` function to determine which label to return.

        :param string language: The preferred language to receive the label in.
            This should be a valid IANA language tag or a list of language tags.
        :rtype: :class:`skosprovider.skos.Label` or None if no labels were found.
        """
        return label(self.labels, language)

    def _sortkey(self, key="id", language="any"):
        """
        Provide a single sortkey for this collection.

        :param string key: Either `id`, `uri`, `label` or `sortlabel`.
        :param string language: The preferred language to receive the label in
            if key is `label` or `sortlabel`. This should be a valid IANA language tag.
        :rtype: :class:`str`
        """
        if key == "id":
            return str(self.id)
        elif key == "uri":
            return self.uri if self.uri else ""
        else:
            sortlabel = label(self.labels, language, key == "sortlabel")
            return sortlabel.label.lower() if sortlabel else ""

    def __repr__(self):
        return f"Concept('{self.id}')"


class Collection:
    """
    A :term:`SKOS` Collection.
    """

    id = None
    """An id for this Collection within a vocabulary"""

    uri = None
    """A proper uri for this Collection"""

    type = "collection"
    """The type of this concept or collection.

    eg. 'collection'
    """

    concept_scheme = None
    """The :class:`ConceptScheme` this Collection is a part of."""

    labels = []
    """A :class:`lst` of :class:`skosprovider.skos.label` instances."""

    notes = []
    """A :class:`lst` of :class:`skosprovider.skos.Note` instances."""

    sources = []
    """A :class:`lst` of :class:`skosprovider.skos.Source` instances."""

    members = []
    """A :class:`lst` of concept or collection ids."""

    member_of = []
    """A :class:`lst` of collection ids."""

    superordinates = []
    """A :class:`lst` of concept ids."""

    infer_concept_relations = True
    """Should member concepts of this collection be seen as narrower concept of
    a superordinate of the collection?"""

    def __init__(
        self,
        id,
        uri=None,
        concept_scheme=None,
        labels=None,
        notes=None,
        sources=None,
        members=None,
        member_of=None,
        superordinates=None,
        infer_concept_relations=True,
    ):
        self.id = id
        self.uri = uri
        self.type = "collection"
        self.concept_scheme = concept_scheme
        self.labels = [dict_to_label(label) for label in labels] if labels else []
        self.notes = [dict_to_note(note) for note in notes] if notes else []
        self.sources = [dict_to_source(source) for source in sources] if sources else []
        self.members = members or []
        self.member_of = member_of or []
        self.superordinates = superordinates or []
        self.infer_concept_relations = infer_concept_relations

    def label(self, language="any"):
        """
        Provide a single label for this collection.

        This uses the :func:`label` function to determine which label to return.

        :param string language: The preferred language to receive the label in.
            This should be a valid IANA language tag.
        :rtype: :class:`skosprovider.skos.Label` or None if no labels were found.
        """
        return label(self.labels, language, False)

    def _sortkey(self, key="id", language="any"):
        """
        Provide a single sortkey for this collection.

        :param string key: Either `id`, `uri`, `label` or `sortlabel`.
        :param string language: The preferred language to receive the label in
            if key is `label` or `sortlabel`. This should be a valid IANA language tag.
        :rtype: :class:`str`
        """
        if key == "id":
            return str(self.id)
        elif key == "uri":
            return self.uri if self.uri else ""
        else:
            sortlabel = label(self.labels, language, key == "sortlabel")
            return sortlabel.label.lower() if sortlabel else ""

    def __repr__(self):
        return f"Collection('{self.id}')"


def label(labels=None, language="any", sortLabel=False):
    """
    Provide a label for a list of labels.

    The items in the list of labels are assumed to be either instances of
    :class:`Label`, or dicts with at least the key `label` in them. These will
    be passed to the :func:`dict_to_label` function.

    This method tries to find a label by looking if there's
    a pref label for the specified language. If there's no pref label,
    it looks for an alt label. It disregards hidden labels.

    While matching languages, preference will be given to exact matches. But,
    if no exact match is present, an inexact match will be attempted. This might
    be because a label in language `nl-BE` is being requested, but only `nl` or
    even `nl-NL` is present. Similarly, when requesting `nl`, a label with
    language `nl-NL` or even `nl-Latn-NL` will also be considered,
    providing no label is present that has an exact match with the
    requested language.

    It's possible to pass multiple languages as a list. In this case, the method
    will try handling each language in turn. Please be aware that this includes
    handling variations. When assing `nl-BE, nl, nl-NL`, the second and third
    languages will never be handled since handling `nl-BE` includes looking for
    other related languages such as `nl-NL` and `nl`.

    If language 'any' was specified, all labels will be considered,
    regardless of language.

    To find a label without a specified language, pass `None` as language.

    If a language or None was specified, and no label could be found, this
    method will automatically try to find a label in some other language.

    Finally, if no label could be found, None is returned.

    ..versionchanged:: 1.1
        It is now possible to pass a list of languages.

    :param any language: The preferred language to receive the label in. This
        should be a valid IANA language tag or list of language tags. If you
        pass a list, the order of the languages in the list will be taken into
        account when trying to determine a label.
    :param boolean sortLabel: Should sortLabels be considered or not? If True,
        sortLabels will be preferred over prefLabels. Bear in mind that these
        are still language dependent. So, it's possible to have a different
        sortLabel per language.
    :rtype: A :class:`Label` or `None` if no label could be found.
    """
    if not labels:
        return None
    if isinstance(language, str):
        language = [language]
    if isinstance(language, list):
        language = [lang for lang in language if tags.tag(lang).language]
    if not language:
        language = ["und"]
    labels = [dict_to_label(label) for label in labels]
    return_label = False
    for lang in language:
        if sortLabel:
            return_label = find_best_label_for_type(labels, lang, "sortLabel")
        if not return_label:
            return_label = find_best_label_for_type(labels, lang, "prefLabel")
        if not return_label:
            return_label = find_best_label_for_type(labels, lang, "altLabel")
        if return_label:
            return return_label
    return label(labels, "any", sortLabel) if "any" not in language else None


def find_best_label_for_type(labels, language, labeltype):
    """
    Find the best label for a certain labeltype.

    :param list labels: A list of :class:`Label`.
    :param str language: An IANA language string, eg. `nl` or `nl-BE`.
    :param str labeltype: Type of label to look for, eg. `prefLabel`.
    """
    typelabels = [label for label in labels if label.type == labeltype]
    if not typelabels:
        return False
    if language == "any":
        return typelabels[0]
    exact = filter_labels_by_language(typelabels, language)
    if exact:
        return exact[0]
    inexact = filter_labels_by_language(typelabels, language, True)
    if inexact:
        return inexact[0]
    return False


def filter_labels_by_language(labels, language, broader=False):
    """
    Filter a list of labels, leaving only labels of a certain language.

    :param list labels: A list of :class:`Label`.
    :param str language: An IANA language string, eg. `nl` or `nl-BE`.
    :param boolean broader: When true, will also match `nl-BE` when filtering
        on `nl`. When false, only exact matches are considered.
    """
    if language == "any":
        return labels
    if broader:
        language = tags.tag(language).language.format
        return [
            label
            for label in labels
            if tags.tag(label.language).language.format == language
        ]
    else:
        language = tags.tag(language).format
        return [
            label for label in labels if tags.tag(label.language).format == language
        ]


def dict_to_label(dict):
    """
    Transform a dict with keys `label`, `type`, `language` and `uri`
    into a :class:`Label`.

    Only the `label` key is mandatory. If `type` is not present, it will
    default to `prefLabel`. If `language` is not present, it will default
    to `und`.

    If the argument passed is not a dict, this method just
    returns the argument.
    """
    try:
        return Label(
            dict["label"],
            dict.get("type", "prefLabel"),
            dict.get("language", "und"),
            uri=dict.get("uri"),
        )
    except (KeyError, AttributeError, TypeError):
        return dict


def dict_to_note(dict):
    """
    Transform a dict with keys `note`, `type` and `language` into a
    :class:`Note`.

    Only the `note` key is mandatory. If `type` is not present, it will
    default to `note`. If `language` is not present, it will default to `und`.
    If `markup` is not present it will default to `None`.

    If the argument passed is already a :class:`Note`, this method just returns
    the argument.
    """
    if isinstance(dict, Note):
        return dict
    return Note(
        dict["note"],
        dict.get("type", "note"),
        dict.get("language", "und"),
        dict.get("markup"),
    )


def dict_to_source(dict):
    """
    Transform a dict with key 'citation' into a :class:`Source`.

    If the argument passed is already a :class:`Source`, this method just
    returns the argument.
    """

    if isinstance(dict, Source):
        return dict
    return Source(dict["citation"], dict.get("markup"))



"""This module provides a registry for skos providers.

This registry helps us find providers during runtime. We can also apply some
operations to all or several providers at the same time.
"""

import logging

from .uri import is_uri


log = logging.getLogger(__name__)


class RegistryException(Exception):
    pass


class Registry:
    """
    This registry collects all skos providers.
    """

    providers = {}
    """
    Dictionary containing all providers, keyed by id.
    """

    concept_scheme_uri_map = {}
    """
    Dictionary mapping concept scheme uri's to vocabulary id's.
    """

    metadata = {}
    """
    Dictionary containing metadata about this registry.
    """

    instance_scope = "single"
    """
    Indicates how the registry is being used. Options:
        - single: The registry is part of a script or a single process. It can
          be assumed to be operational for the entire duration of the process
          and there are no threads involved.
        - threaded_global: The registry is part of a program that uses threads,
          such as a typical web application. It's attached to the global process
          and duplicated to threads, making it not thread safe. Proceed carefully
          with certain providers. Should generally only be used with
          applications that only use read-only providers that load all data in
          memory at startup and use no database connections or other kinds of
          sessions.
        - threaded_thread: The registry is part of a program that uses threads,
          such as a typical web application. It's attached to a thread, such as
          a web request. The registry is instantiated for this thread/request and
          dies with this thread/request. This is needed for providers such
          as the SQLAlchemyProvider. Providers that use database connections or
          other session handling code generally require this.
    """

    def __init__(self, instance_scope="single", metadata=None):
        """
        :param str instance_scope: Indicates how the registry was instantiated.
            Possible values: single, threaded_global, threaded_thread.
        :param dict metadata: Metadata essential to this registry. Possible
            metadata:

                * `catalog`: A :class:`dict` detailing the catalog all \
                    conceptschemes are part of. \
                    Currently the contents of the dictionary are undefined \
                    except for a :term:`uri` attribute that must be present.
                * `dataset`: A :class:`dict` detailing the dataset all \
                    conceptschemes are part of. \
                    Currently the contents of the dictionary are undefined \
                    except for a :term:`uri` attribute that must be present.
        """
        self.providers = {}
        self.concept_scheme_uri_map = {}
        self.metadata = metadata or {}
        if instance_scope not in ["single", "threaded_global", "threaded_thread"]:
            raise ValueError("Invalid instance_scope.")
        self.instance_scope = instance_scope

    def get_metadata(self):
        """Get some metadata on the registry it represents.

        :rtype: Dict.
        """
        return self.metadata

    def register_provider(self, provider):
        """
        Register a :class:`skosprovider.providers.VocabularyProvider`.

        :param skosprovider.providers.VocabularyProvider provider: The provider
            to register.
        :raises RegistryException: A provider with this id or uri has already
            been registered.
        """
        if (
            provider.allowed_instance_scopes
            and self.instance_scope not in provider.allowed_instance_scopes
        ):
            raise RegistryException(
                f"This provider does not support instance_scope {self.instance_scope}"
            )
        if provider.get_vocabulary_id() in self.providers:
            raise RegistryException(
                "A provider with this id has already been registered."
            )
        self.providers[provider.get_vocabulary_id()] = provider
        try:
            conceptscheme_uri = provider.get_vocabulary_uri()
        except AttributeError as e:
            log.error(e)
            # For providers not compatible with skosprovider >= 0.8.0
            log.warning(
                "New versions of skosprovider (>=0.8.0) require your provider "
                "to have a get_vocabulary_uri method. This fallback mechanism "
                "will be removed in version 2.0.0."
            )
            conceptscheme_uri = provider.concept_scheme.uri
        if conceptscheme_uri in self.concept_scheme_uri_map:
            raise RegistryException(
                "A provider with URI {conceptscheme_uri} has already been registered."
            )
        self.concept_scheme_uri_map[conceptscheme_uri] = provider.get_vocabulary_id()

    def remove_provider(self, id):
        """
        Remove the provider with the given id or :term:`URI`.

        :param str id: The identifier for the provider.
        :returns: A :class:`skosprovider.providers.VocabularyProvider` or
            `False` if the id is unknown.
        """
        if id in self.providers:
            p = self.providers.get(id, False)
            del self.providers[id]
            try:
                cs_uri = p.get_vocabulary_uri()
            except AttributeError as e:
                log.error(e)
                # For providers not compatible with skosprovider >= 0.8.0
                log.warning(
                    "New versions of skosprovider (>=0.8.0) require your provider "
                    "to have a get_vocabulary_uri method. This fallback mechanism "
                    "will be removed in version 2.0.0."
                )
                # For providers not compatible with skosprovider >= 0.8.0
                cs_uri = p.concept_scheme.uri
            del self.concept_scheme_uri_map[cs_uri]
            return p
        elif id in self.concept_scheme_uri_map:
            id = self.concept_scheme_uri_map[id]
            return self.remove_provider(id)
        else:
            return False

    def get_provider(self, id):
        """
        Get a provider by id or :term:`URI`.

        :param str id: The identifier for the provider. This can either be the
            id with which it was registered or the :term:`URI` of the conceptscheme
            that the provider services.
        :returns: A :class:`skosprovider.providers.VocabularyProvider`
            or `False` if the id or uri is unknown.
        """
        if id in self.providers:
            return self.providers.get(id, False)
        elif is_uri(id) and id in self.concept_scheme_uri_map:
            return self.providers.get(self.concept_scheme_uri_map[id], False)
        return False

    def get_providers(self, **kwargs):
        """Get all providers registered.

        If keyword `ids` is present, get only the providers with these ids.

        If keys `subject` is present, get only the providers that have this subject.

        .. code-block:: python

           # Get all providers with subject 'biology'
           registry.get_providers(subject='biology')

           # Get all providers with id 1 or 2
           registry.get_providers(ids=[1,2])

           # Get all providers with id 1 or 2 and subject 'biology'
           registry.get_providers(ids=[1,2], subject='biology']

        :param list ids: Only return providers with one of the Ids
            or :term:`URIs <URI>`.
        :param str subject: Only return providers with this subject.
        :returns: A list of
            :class:`providers <skosprovider.providers.VocabularyProvider>`
        """
        if "ids" in kwargs:
            ids = [self.concept_scheme_uri_map.get(id, id) for id in kwargs["ids"]]
            providers = [self.providers[k] for k in self.providers.keys() if k in ids]
        else:
            providers = list(self.providers.values())
        if "subject" in kwargs:
            providers = [
                p for p in providers if kwargs["subject"] in p.metadata["subject"]
            ]
        return providers

    def find(self, query, **kwargs):
        """Launch a query across all or a selection of providers.

        .. code-block:: python

            # Find anything that has a label of church in any provider.
            registry.find({'label': 'church'})

            # Find anything that has a label of church with the BUILDINGS provider.
            # Attention, this syntax was deprecated in version 0.3.0
            registry.find({'label': 'church'}, providers=['BUILDINGS'])

            # Find anything that has a label of church with the BUILDINGS provider.
            registry.find({'label': 'church'}, providers={'ids': ['BUILDINGS']})

            # Find anything that has a label of church with a provider
            # marked with the subject 'architecture'.
            registry.find({'label': 'church'}, providers={'subject': 'architecture'})

            # Find anything that has a label of church in any provider.
            # If possible, display the results with a Dutch label.
            registry.find({'label': 'church'}, language='nl')

            # Find anything that has a match with an external concept
            # If possible, display the results with a Dutch label.
            registry.find({
                'matches': {
                    'uri': 'http://id.py.org/diff/types/of/trees/nr/1/the/larch'
                }}, language='nl')

            # Find anything that has a label of lariks
            # with a close match to an external concept
            # If possible, display the results with a Dutch label.
            provider.find({
                'matches': {
                    'label': 'lariks',
                    'type': 'close',
                    'uri': 'http://id.py.org/diff/types/of/trees/nr/1/the/larch'
                }}, language='nl')

        :param dict query: The query parameters that will be passed on to each
            :meth:`~skosprovider.providers.VocabularyProvider.find` method of
            the selected.
            :class:`providers <skosprovider.providers.VocabularyProvider>`.
        :param dict providers: Optional. If present, it should be a dictionary.
            This dictionary can contain any of the keyword arguments available
            to the :meth:`get_providers` method. The query will then only
            be passed to the providers confirming to these arguments.
        :param string language: Optional. If present, it should be a
            :term:`language-tag`. This language-tag is passed on to the
            underlying providers and used when selecting the label to display
            for each concept.
        :returns: a list of :class:`dict`.
            Each dict has two keys: id and concepts.
        """
        if "providers" not in kwargs:
            providers = self.get_providers()
        else:
            pargs = kwargs["providers"]
            if isinstance(pargs, list):
                providers = self.get_providers(ids=pargs)
            else:
                providers = self.get_providers(**pargs)
        kwarguments = {}
        if "language" in kwargs:
            kwarguments["language"] = kwargs["language"]
        return [
            {"id": p.get_vocabulary_id(), "concepts": p.find(query, **kwarguments)}
            for p in providers
        ]

    def get_all(self, **kwargs):
        """Get all concepts from all providers.

        .. code-block:: python

            # get all concepts in all providers.
            registry.get_all()

            # get all concepts in all providers.
            # If possible, display the results with a Dutch label.
            registry.get_all(language='nl')

        :param string language: Optional. If present, it should be a
            :term:`language-tag`. This language-tag is passed on to the
            underlying providers and used when selecting the label to display
            for each concept.

        :returns: a list of :class:`dict`.
            Each dict has two keys: id and concepts.
        """
        kwarguments = {}
        if "language" in kwargs:
            kwarguments["language"] = kwargs["language"]
        return [
            {"id": p.get_vocabulary_id(), "concepts": p.get_all(**kwarguments)}
            for p in self.providers.values()
        ]

    def get_by_uri(self, uri):
        """Get a concept or collection by its uri.

        Returns a single concept or collection if one exists with this uri.
        Returns False otherwise.

        :param string uri: The uri to find a concept or collection for.
        :raises ValueError: The uri is invalid.
        :rtype: :class:`skosprovider.skos.Concept` or
            :class:`skosprovider.skos.Collection`
        """
        if not is_uri(uri):
            raise ValueError(f"{uri} is not a valid URI.")
        # Check if there's a provider that's more likely to have the URI
        csuris = [
            csuri
            for csuri in self.concept_scheme_uri_map.keys()
            if uri.startswith(csuri)
        ]
        for csuri in csuris:
            c = self.get_provider(csuri).get_by_uri(uri)
            if c:
                return c
        # Check all providers
        for p in self.providers.values():
            c = p.get_by_uri(uri)
            if c:
                return c
        return False



"""
This module contains functions dealing with jsonld reading and writing.

.. versionadded:: 0.7.0
"""

import logging

from skosprovider.utils import add_lang_to_html
from skosprovider.utils import extract_language

log = logging.getLogger(__name__)

MINI_CONTEXT = {
    "@version": 1.1,
    "dct": "http://purl.org/dc/terms/",
    "skos": "http://www.w3.org/2004/02/skos/core#",
    "iso-thes": "http://purl.org/iso25964/skos-thes#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "void": "http://rdfs.org/ns/void#",
    "uri": "@id",
    "type": "@type",
    "id": "dct:identifier",
    "label": "rdfs:label",
    "concept": "skos:Concept",
    "collection": "skos:Collection",
}

CONTEXT = {
    "@version": 1.1,
    "dct": "http://purl.org/dc/terms/",
    "skos": "http://www.w3.org/2004/02/skos/core#",
    "skosxl": "http://www.w3.org/2008/05/skos-xl#",
    "iso-thes": "http://purl.org/iso25964/skos-thes#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "void": "http://rdfs.org/ns/void#",
    "uri": "@id",
    "type": "@type",
    "id": "dct:identifier",
    "label": "rdfs:label",
    "concept": "skos:Concept",
    "collection": "skos:Collection",
    "lbl": "@value",
    "nt": "@value",
    "ct": "@value",
    "HTML": "rdf:HTML",
    "concept_scheme": {"@id": "skos:inScheme", "@type": "@id"},
    "in_dataset": {"@id": "void:inDataset", "@type": "@id"},
    "top_concepts": {"@id": "skos:hasTopConcept", "@type": "@id"},
    "labels": "@nest",
    "labels_xl": "@nest",
    "notes": "@nest",
    "matches": "@nest",
    "sources": {"@id": "dct:source", "@container": "@set"},
    "citations": {"@id": "dct:bibliographicCitation", "@container": "@set"},
    "pref_labels": {"@nest": "labels", "@id": "skos:prefLabel", "@container": "@set"},
    "alt_labels": {"@nest": "labels", "@id": "skos:altLabel", "@container": "@set"},
    "hidden_labels": {
        "@nest": "labels",
        "@id": "skos:hiddenLabel",
        "@container": "@set",
    },
    "pref_labels_xl": {
        "@nest": "labels_xl",
        "@id": "skosxl:prefLabel",
        "@container": "@set",
    },
    "alt_labels_xl": {
        "@nest": "labels_xl",
        "@id": "skosxl:altLabel",
        "@container": "@set",
    },
    "hidden_labels_xl": {
        "@nest": "labels_xl",
        "@id": "skosxl:hiddenLabel",
        "@container": "@set",
    },
    "general_notes": {"@nest": "notes", "@id": "skos:note", "@container": "@set"},
    "scope_notes": {"@nest": "notes", "@id": "skos:scopeNote", "@container": "@set"},
    "definitions": {"@nest": "notes", "@id": "skos:definition", "@container": "@set"},
    "history_notes": {
        "@nest": "notes",
        "@id": "skos:historyNote",
        "@container": "@set",
    },
    "editorial_notes": {
        "@nest": "notes",
        "@id": "skos:editorialNote",
        "@container": "@set",
    },
    "change_notes": {"@nest": "notes", "@id": "skos:changeNote", "@container": "@set"},
    "examples": {"@nest": "notes", "@id": "skos:example", "@container": "@set"},
    "exact_matches": {"@id": "skos:exactMatch", "@type": "@id", "@container": "@set"},
    "close_matches": {"@id": "skos:closeMatch", "@type": "@id", "@container": "@set"},
    "broad_matches": {"@id": "skos:broadMatch", "@type": "@id", "@container": "@set"},
    "narrow_matches": {"@id": "skos:narrowMatch", "@type": "@id", "@container": "@set"},
    "related_matches": {
        "@id": "skos:relatedMatch",
        "@type": "@id",
        "@container": "@set",
    },
    "member_of": {"@reverse": "skos:member", "@type": "@id", "@container": "@set"},
    "members": {"@id": "skos:member", "@type": "@id", "@container": "@set"},
    "subordinate_arrays": {
        "@id": "iso-thes:subordinateArray",
        "@type": "@id",
        "@container": "@set",
    },
    "broader": {"@id": "skos:broader", "@type": "@id", "@container": "@set"},
    "narrower": {"@id": "skos:narrower", "@type": "@id", "@container": "@set"},
    "related": {"@id": "skos:related", "@type": "@id", "@container": "@set"},
    "languages": {"@id": "dct:language", "@type": "@id", "@container": "@set"},
}


def jsonld_dumper(provider, context=None, language=None):
    """
    Dump a provider to a JSON-LD serialisable dictionary.

    :param skosprovider.providers.VocabularyProvider provider: The provider
        that wil be turned into a JSON-LD `dict`.
    :param str or dict context: Context as a dict or link to context file.
    :param string language: Language to render a single label in.

    :rtype: A `dict`
    """
    if not language:
        language = provider.metadata.get("default_language", "en")
    doc = {"@graph": []}
    if context:
        doc["@context"] = context
    doc["@graph"].append(
        jsonld_conceptscheme_dumper(
            provider, None, relations_profile="uri", language=language
        )
    )
    for c in provider.get_all():
        doc["@graph"].append(
            jsonld_c_dumper(
                provider, c["id"], None, relations_profile="uri", language=language
            )
        )
    return doc


def jsonld_c_dumper(
    provider, id, context=None, relations_profile="partial", language="en"
):
    """
    Dump a concept or collection to a JSON-LD serialisable dictionary.

    :param skosprovider.providers.VocabularyProvider provider: The provider
        that contains the concept or collection.
    :param str or int id: Identifier of the concept to dump.
    :param str or dict context: Context as a dict or link to context file.
    :param str relations_profile: Either `partial` or `uri` to render links to
        other resources with some information or just a :term:`URI`.
    :param string language: Language to render a single label in.

    :rtype: A `dict`
    """
    c = provider.get_by_id(id)
    doc = _jsonld_c_basic_renderer(c, language)
    if context:
        doc["@context"] = context
    if relations_profile == "partial":
        doc["concept_scheme"] = _jsonld_cs_basic_renderer(c.concept_scheme, language)
    else:
        doc["concept_scheme"] = c.concept_scheme.uri
    dataset_uri = provider.get_metadata().get("dataset", {}).get("uri", None)
    if dataset_uri:
        doc["in_dataset"] = dataset_uri
    doc.update(_jsonld_labels_renderer(c))
    doc.update(_jsonld_labels_xl_renderer(c))
    doc.update(_jsonld_notes_renderer(c))
    doc.update(_jsonld_sources_renderer(c))
    doc.update(_jsonld_member_of_renderer(c, provider, relations_profile, language))
    if c.type == "concept":
        doc.update(_jsonld_matches_renderer(c))
        doc.update(_jsonld_broader_renderer(c, provider, relations_profile, language))
        doc.update(_jsonld_narrower_renderer(c, provider, relations_profile, language))
        doc.update(_jsonld_related_renderer(c, provider, relations_profile, language))
        doc.update(
            _jsonld_subordinate_arrays_renderer(
                c, provider, relations_profile, language
            )
        )
    elif c.type == "collection":
        doc["infer_concept_relations"] = True
        doc.update(_jsonld_members_renderer(c, provider, relations_profile, language))
        doc.update(
            _jsonld_superordinates_renderer(c, provider, relations_profile, language)
        )
    return doc


def _jsonld_c_basic_renderer(c, language="en"):
    doc = {"id": c.id, "uri": c.uri, "type": c.type}
    label = c.label(language)
    if label:
        doc["label"] = label.label
    return doc


def _jsonld_cs_basic_renderer(cs, language="en"):
    doc = {"uri": cs.uri, "type": "skos:ConceptScheme"}
    label = cs.label(language)
    if label:
        doc["label"] = label.label
    return doc


def _jsonld_labels_renderer(c):
    if not len(c.labels):
        return {}
    doc = {"labels": {}}

    def lbl_renderer(label):
        language = extract_language(label.language)
        return {"language": language, "@language": language, "lbl": label.label}

    ltypemap = {
        "prefLabel": "pref_labels",
        "altLabel": "alt_labels",
        "hiddenLabel": "hidden_labels",
        "sortLabel": "hidden_labels",
    }
    for label in c.labels:
        doc["labels"].setdefault(ltypemap[label.type], []).append(lbl_renderer(label))
    return doc


def _jsonld_labels_xl_renderer(c):
    if not len([label for label in c.labels if label.is_xl()]):
        return {}
    doc = {"labels_xl": {}}

    def lbl_xl_renderer(label):
        language = extract_language(label.language)
        return {
            "uri": label.uri,
            "type": "skosxl:Label",
            "skosxl:literalForm": {"@language": language, "lbl": label.label},
        }

    ltypemap = {
        "prefLabel": "pref_labels_xl",
        "altLabel": "alt_labels_xl",
        "hiddenLabel": "hidden_labels_xl",
        "sortLabel": "hidden_labels_xl",
    }
    for label in c.labels:
        if label.is_xl():
            doc["labels_xl"].setdefault(ltypemap[label.type], []).append(
                lbl_xl_renderer(label)
            )
    return doc


def _jsonld_notes_renderer(c):
    if not len(c.notes):
        return {}
    doc = {"notes": {}}

    def nt_renderer(n):
        language = extract_language(n.language)
        note = {"language": language, "@language": language, "nt": n.note}
        if n.markup is not None:
            del note["@language"]
            note["nt"] = add_lang_to_html(note["nt"], language)
            note["@type"] = n.markup
        return note

    ntypemap = {
        "note": "general_notes",
        "scopeNote": "scope_notes",
        "definition": "definitions",
        "historyNote": "history_notes",
        "editorialNote": "editorial_notes",
        "changeNote": "change_notes",
        "example": "examples",
    }
    for n in c.notes:
        doc["notes"].setdefault(ntypemap[n.type], []).append(nt_renderer(n))
    return doc


def _jsonld_sources_renderer(c):
    if not len(c.sources):
        return {}
    doc = {"sources": []}

    def s_renderer(s):
        source = {
            "type": "dct:BibliographicResource",
            "citations": [{"ct": s.citation}],
        }
        if s.markup is not None:
            source["citations"][0]["@type"] = s.markup
        return source

    for s in c.sources:
        doc["sources"].append(s_renderer(s))
    return doc


def _jsonld_matches_renderer(c):
    if not any([len(matches) for matches in c.matches.values()]):
        return {}
    doc = {"matches": {}}
    for matchtype, matches in c.matches.items():
        if len(matches):
            doc["matches"].setdefault(f"{matchtype}_matches", []).extend(matches)
    return doc


def _jsonld_superordinates_renderer(c, provider, profile="partial", language="en"):
    return _jsonld_relation_renderer(c, provider, "superordinates", profile, language)


def _jsonld_members_renderer(c, provider, profile="partial", language="en"):
    return _jsonld_relation_renderer(c, provider, "members", profile, language)


def _jsonld_member_of_renderer(c, provider, profile="partial", language="en"):
    return _jsonld_relation_renderer(c, provider, "member_of", profile, language)


def _jsonld_broader_renderer(c, provider, profile="partial", language="en"):
    return _jsonld_relation_renderer(c, provider, "broader", profile, language)


def _jsonld_narrower_renderer(c, provider, profile="partial", language="en"):
    return _jsonld_relation_renderer(c, provider, "narrower", profile, language)


def _jsonld_related_renderer(c, provider, profile="partial", language="en"):
    return _jsonld_relation_renderer(c, provider, "related", profile, language)


def _jsonld_subordinate_arrays_renderer(c, provider, profile="partial", language="en"):
    return _jsonld_relation_renderer(
        c, provider, "subordinate_arrays", profile, language
    )


def _jsonld_relation_renderer(c, provider, relation, profile="partial", language="en"):
    doc = {relation: []}
    for m in getattr(c, relation):
        relc = provider.get_by_id(m)
        if profile == "partial":
            doc[relation].append(_jsonld_c_basic_renderer(relc, language))
        else:
            doc[relation].append(relc.uri)
    return doc


def _jsonld_topconcepts_renderer(provider, profile="partial"):
    doc = {"top_concepts": []}
    for c in provider.get_top_concepts():
        if profile == "partial":
            doc["top_concepts"].append(c)
        else:
            doc["top_concepts"].append(c["uri"])
    return doc


def _jsonld_cs_languages_renderer(cs):
    doc = {"languages": []}
    for language in cs.languages:
        doc["languages"].append(language)
    return doc


def jsonld_conceptscheme_dumper(
    provider, context=None, relations_profile="partial", language="en"
):
    """
    Dump a conceptscheme to a JSON-LD serialisable dictionary.

    :param skosprovider.providers.VocabularyProvider provider: The provider
        that contains the conceptscheme.
    :param str or dict context: Context as a dict or link to context file.
    :param str relations_profile: Either `partial` or `uri` to render links to
        other resources with some information or just a :term:`URI`.
    :param string language: Language to render a single label in.

    :rtype: A `dict`
    """
    conceptscheme = provider.concept_scheme
    doc = _jsonld_cs_basic_renderer(conceptscheme, language)
    if context:
        doc["@context"] = context
    dataset_uri = provider.get_metadata().get("dataset", {}).get("uri", None)
    if dataset_uri:
        doc["in_dataset"] = dataset_uri
    doc["id"] = provider.get_metadata()["id"]
    doc.update(_jsonld_labels_renderer(conceptscheme))
    doc.update(_jsonld_labels_xl_renderer(conceptscheme))
    doc.update(_jsonld_notes_renderer(conceptscheme))
    doc.update(_jsonld_sources_renderer(conceptscheme))
    doc.update(_jsonld_cs_languages_renderer(conceptscheme))
    doc.update(_jsonld_topconcepts_renderer(provider, relations_profile))
    return doc



import unittest

from skosprovider.uri import DefaultConceptSchemeUrnGenerator
from skosprovider.uri import DefaultUrnGenerator
from skosprovider.uri import TypedUrnGenerator
from skosprovider.uri import UriPatternGenerator
from skosprovider.uri import is_uri


class IsUriTest(unittest.TestCase):

    def test_None(self):
        assert not is_uri(None)

    def test_url(self):
        assert is_uri("https://id.erfgoed.net/thesauri/erfgoedtypes/1")
        assert is_uri("https://thesaurus.erfgoed.net/conceptschemes/erfgoedtypes/1")

    def test_urn(self):
        assert is_uri("urn:x-skosprovider:typologie")
        assert is_uri("urn:x-skosprovider:typologie:1")


class UriPatternGeneratorTest(unittest.TestCase):

    def test_simple(self):
        urigen = UriPatternGenerator("http://id.example.com/%s")
        self.assertEqual("http://id.example.com/1", urigen.generate(id=1))


class DefaultUrnGeneratorTest(unittest.TestCase):

    def setUp(self):
        self.urigen = DefaultUrnGenerator("typologie")

    def tearDown(self):
        del self.urigen

    def test_simple(self):
        self.assertEqual("urn:x-skosprovider:typologie:1", self.urigen.generate(id=1))

    def test_missing_argument(self):
        self.assertRaises(KeyError, self.urigen.generate, type="set")


class DefaultConceptSchemeUrnGeneratorTest(unittest.TestCase):

    def setUp(self):
        self.urigen = DefaultConceptSchemeUrnGenerator()

    def tearDown(self):
        del self.urigen

    def test_simple(self):
        self.assertEqual(
            "urn:x-skosprovider:typologie", self.urigen.generate(id="TYPOLOGIE")
        )

    def test_missing_argument(self):
        self.assertRaises(KeyError, self.urigen.generate)


class TypedUrnGeneratorTest(unittest.TestCase):

    def setUp(self):
        self.urigen = TypedUrnGenerator("typologie")

    def tearDown(self):
        del self.urigen

    def test_concept(self):
        self.assertEqual(
            "urn:x-skosprovider:typologie:concept:1",
            self.urigen.generate(type="concept", id=1),
        )

    def test_collection(self):
        self.urigen = TypedUrnGenerator("typologie")
        self.assertEqual(
            "urn:x-skosprovider:typologie:collection:7000",
            self.urigen.generate(type="collection", id=7000),
        )

    def test_invalid_type(self):
        self.assertRaises(ValueError, self.urigen.generate, type="set", id=1)



from test_providers import geo
from test_providers import larch
from test_providers import trees

from skosprovider.jsonld import CONTEXT
from skosprovider.jsonld import jsonld_c_dumper
from skosprovider.jsonld import jsonld_conceptscheme_dumper
from skosprovider.jsonld import jsonld_dumper


class TestDumperTrees:

    def test_dump_larch(self):
        doc = jsonld_c_dumper(trees, 1, language="nl")
        assert doc["id"] == "1"
        assert doc["uri"] == "http://id.trees.org/1"
        assert doc["type"] == "concept"
        assert doc["label"] == "De Lariks"
        assert len(doc["labels"]["pref_labels"]) == 2
        assert {"language": "en", "@language": "en", "lbl": "The Larch"} in doc[
            "labels"
        ]["pref_labels"]
        assert len(doc["notes"]["definitions"]) == 1
        assert {
            "language": "en",
            "@language": "en",
            "nt": "A type of tree.",
        } in doc[
            "notes"
        ]["definitions"]
        assert len(doc["sources"]) == len(larch["sources"])
        assert {
            "type": "dct:BibliographicResource",
            "citations": [
                {
                    "ct": "Monthy Python. Episode Three: How to recognise different"
                    " types of trees from quite a long way away."
                }
            ],
        } in doc["sources"]
        assert len(doc["member_of"]) == 1
        assert (
            "http://id.python.org/different/types/of/trees/nr/1/the/larch"
            in doc["matches"]["exact_matches"]
        )
        assert doc["concept_scheme"] == {
            "uri": "http://id.trees.org",
            "type": "skos:ConceptScheme",
            "label": "Soorten",
        }
        assert doc["in_dataset"] == "http://id.trees.org/dataset"

    def test_dump_larch_uri_profile(self):
        doc = jsonld_c_dumper(trees, 1, relations_profile="uri")
        assert doc["member_of"] == ["http://id.trees.org/3"]
        assert doc["concept_scheme"] == "http://id.trees.org"

    def test_dump_larch_label_en(self):
        doc = jsonld_c_dumper(trees, 1, language="en")
        assert doc["label"] == "The Larch"

    def test_dump_larch_label_nl(self):
        doc = jsonld_c_dumper(trees, 1, language="nl")
        assert doc["label"] == "De Lariks"

    def test_dump_larch_partial_profile_en(self):
        doc = jsonld_c_dumper(trees, 1, relations_profile="partial", language="en")
        assert doc["member_of"] == [
            {
                "id": 3,
                "uri": "http://id.trees.org/3",
                "type": "collection",
                "label": "Trees by species",
            }
        ]
        assert doc["concept_scheme"] == {
            "uri": "http://id.trees.org",
            "type": "skos:ConceptScheme",
            "label": "Species",
        }

    def test_dump_larch_inline_context(self):
        doc = jsonld_c_dumper(trees, 1, CONTEXT)
        assert "@context" in doc
        assert doc["@context"] == CONTEXT

    def test_dump_chestnut(self):
        doc = jsonld_c_dumper(trees, 2)
        assert doc["id"] == "2"
        assert doc["uri"] == "http://id.trees.org/2"
        assert doc["type"] == "concept"
        assert doc["label"] == "The Chestnut"
        assert len(doc["labels"]["pref_labels"]) == 1
        assert {"language": "en", "@language": "en", "lbl": "The Chestnut"} in doc[
            "labels"
        ]["pref_labels"]
        assert len(doc["labels"]["alt_labels"]) == 2
        assert "labels_xl" not in doc
        assert len(doc["notes"]["definitions"]) == 1
        assert {
            "language": "en",
            "@language": "en",
            "nt": "A different type of tree.",
        } in doc["notes"]["definitions"]
        assert len(doc["member_of"]) == 1
        assert len(doc["matches"]["related_matches"]) == 1
        assert doc["concept_scheme"] == {
            "uri": "http://id.trees.org",
            "type": "skos:ConceptScheme",
            "label": "Species",
        }
        assert doc["in_dataset"] == "http://id.trees.org/dataset"

    def test_dump_species(self):
        doc = jsonld_c_dumper(trees, 3)
        assert doc["id"] == 3
        assert doc["uri"] == "http://id.trees.org/3"
        assert doc["type"] == "collection"
        assert doc["label"] == "Trees by species"
        assert len(doc["labels"]["pref_labels"]) == 2
        assert {"language": "en", "@language": "en", "lbl": "Trees by species"} in doc[
            "labels"
        ]["pref_labels"]
        assert len(doc["labels"]["hidden_labels"]) == 1
        assert "labels_xl" not in doc
        assert len(doc["notes"]["editorial_notes"]) == 1
        assert {
            "language": "en",
            "nt": '<div xml:lang="en">As seen in <em>How to Recognise '
            "Different Types of Trees from Quite a Long Way Away</em>.</div>",
            "@type": "HTML",
        } in doc["notes"]["editorial_notes"]
        assert "sources" not in doc
        assert len(doc["members"]) == 2
        assert {
            "id": "2",
            "uri": "http://id.trees.org/2",
            "type": "concept",
            "label": "The Chestnut",
        } in doc["members"]
        assert "matches" not in doc
        assert doc["concept_scheme"] == {
            "uri": "http://id.trees.org",
            "type": "skos:ConceptScheme",
            "label": "Species",
        }
        assert doc["in_dataset"] == "http://id.trees.org/dataset"

    def test_dump_trees_cs_nl(self):
        doc = jsonld_conceptscheme_dumper(trees, language="nl")
        assert doc["uri"] == "http://id.trees.org"
        assert doc["type"] == "skos:ConceptScheme"
        assert doc["id"] == "TREES"
        assert doc["label"] == "Soorten"
        assert len(doc["top_concepts"]) == 2
        assert len(doc["labels"]["pref_labels"]) == 2
        assert len(doc["labels_xl"]["pref_labels_xl"]) == 1
        assert "sources" not in doc
        assert "notes" not in doc
        assert doc["in_dataset"] == "http://id.trees.org/dataset"

    def test_dump_trees_cs_partial_profile(self):
        doc = jsonld_conceptscheme_dumper(
            trees, relations_profile="partial", language="nl"
        )
        assert len(doc["top_concepts"]) == 2
        assert {
            "id": "2",
            "uri": "http://id.trees.org/2",
            "type": "concept",
            "label": "De Paardekastanje",
        } in doc["top_concepts"]

    def test_dump_trees_cs_uri_profile(self):
        doc = jsonld_conceptscheme_dumper(trees, relations_profile="uri")
        assert len(doc["top_concepts"]) == 2
        assert "http://id.trees.org/1" in doc["top_concepts"]
        assert "http://id.trees.org/2" in doc["top_concepts"]

    def test_dump_trees_cs_inline_context(self):
        doc = jsonld_conceptscheme_dumper(trees, CONTEXT)
        assert doc["uri"] == "http://id.trees.org"
        assert doc["type"] == "skos:ConceptScheme"
        assert "@context" in doc
        assert doc["@context"] == CONTEXT

    def test_dump_trees_cs_url_context(self):
        context_uri = "https://atramhasis.org/context/atramhasis.jsonld"
        doc = jsonld_conceptscheme_dumper(trees, context_uri)
        assert doc["uri"] == "http://id.trees.org"
        assert doc["type"] == "skos:ConceptScheme"
        assert "@context" in doc
        assert doc["@context"] == context_uri

    def test_dump_trees_cs_xllabel(self):
        doc = jsonld_conceptscheme_dumper(trees)
        assert doc["label"] == "Species"
        assert "labels_xl" in doc
        assert "pref_labels_xl" in doc["labels_xl"]
        assert len(doc["labels_xl"]["pref_labels_xl"]) == 1
        assert doc["labels_xl"]["pref_labels_xl"][0]["type"] == "skosxl:Label"
        assert (
            doc["labels_xl"]["pref_labels_xl"][0]["uri"]
            == "http://id.trees.org/labels/soorten-nl"
        )
        assert doc["labels_xl"]["pref_labels_xl"][0]["skosxl:literalForm"] == {
            "@language": "nl",
            "lbl": "Soorten",
        }
        assert {"language": "nl", "@language": "nl", "lbl": "Soorten"} in doc["labels"][
            "pref_labels"
        ]

    def test_dump_trees(self):
        doc = jsonld_dumper(trees)
        assert "@graph" in doc
        assert len(doc["@graph"]) == 4

    def test_dump_trees_inline_context(self):
        doc = jsonld_dumper(trees, CONTEXT)
        assert "@graph" in doc
        assert len(doc["@graph"]) == 4
        assert "@context" in doc
        assert doc["@context"] == CONTEXT

    def test_dump_trees_url_context(self):
        context_uri = "https://atramhasis.org/context/atramhasis.jsonld"
        doc = jsonld_dumper(trees, context_uri)
        assert "@graph" in doc
        assert len(doc["@graph"]) == 4
        assert "@context" in doc
        assert doc["@context"] == context_uri


class TestDumperGeo:

    def test_dump_geo(self):
        doc = jsonld_dumper(geo, CONTEXT)
        assert "@graph" in doc
        assert len(doc["@graph"]) == 20
        assert "@context" in doc
        assert doc["@context"] == CONTEXT

    def test_dump_Belgium(self):
        doc = jsonld_c_dumper(geo, 4, CONTEXT)
        assert len(doc["subordinate_arrays"]) == 2
        assert "matches" not in doc
